import copy
from typing import Any
from typing import Literal
from typing import Union

from pydantic import model_validator

from laktory._logger import get_logger
from laktory._parsers import merge_dicts
from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.pipeline.pipeline import Pipeline
from laktory.models.resources.baseresource import ResourceOptions
from laktory.models.resources.databricks.alert import Alert
from laktory.models.resources.databricks.catalog import Catalog
from laktory.models.resources.databricks.cluster import Cluster
from laktory.models.resources.databricks.clusterpolicy import ClusterPolicy
from laktory.models.resources.databricks.dashboard import Dashboard
from laktory.models.resources.databricks.dbfsfile import DbfsFile
from laktory.models.resources.databricks.directory import Directory
from laktory.models.resources.databricks.dltpipeline import DLTPipeline
from laktory.models.resources.databricks.externallocation import ExternalLocation
from laktory.models.resources.databricks.grant import Grant
from laktory.models.resources.databricks.grants import Grants
from laktory.models.resources.databricks.group import Group
from laktory.models.resources.databricks.job import Job
from laktory.models.resources.databricks.metastore import Metastore
from laktory.models.resources.databricks.metastoreassignment import MetastoreAssignment
from laktory.models.resources.databricks.metastoredataaccess import MetastoreDataAccess
from laktory.models.resources.databricks.mlflowexperiment import MLflowExperiment
from laktory.models.resources.databricks.mlflowmodel import MLflowModel
from laktory.models.resources.databricks.mlflowwebhook import MLflowWebhook
from laktory.models.resources.databricks.mwsnetworkconnectivityconfig import (
    MwsNetworkConnectivityConfig,
)
from laktory.models.resources.databricks.notebook import Notebook
from laktory.models.resources.databricks.query import Query
from laktory.models.resources.databricks.repo import Repo
from laktory.models.resources.databricks.schema import Schema
from laktory.models.resources.databricks.secret import Secret
from laktory.models.resources.databricks.secretscope import SecretScope
from laktory.models.resources.databricks.serviceprincipal import ServicePrincipal
from laktory.models.resources.databricks.table import Table
from laktory.models.resources.databricks.user import User
from laktory.models.resources.databricks.vectorsearchendpoint import (
    VectorSearchEndpoint,
)
from laktory.models.resources.databricks.vectorsearchindex import VectorSearchIndex
from laktory.models.resources.databricks.volume import Volume
from laktory.models.resources.databricks.warehouse import Warehouse
from laktory.models.resources.databricks.workspacebinding import WorkspaceBinding
from laktory.models.resources.databricks.workspacefile import WorkspaceFile
from laktory.models.resources.providers.awsprovider import AWSProvider
from laktory.models.resources.providers.azureprovider import AzureProvider
from laktory.models.resources.providers.azurepulumiprovider import AzurePulumiProvider
from laktory.models.resources.providers.baseprovider import BaseProvider
from laktory.models.resources.providers.databricksprovider import DatabricksProvider

logger = get_logger(__name__)

DIRPATH = "./"


class Terraform(BaseModel):
    backend: Union[dict[str, Any], None] = None


class LaktorySettings(BaseModel):
    """
    Laktory Settings

    Attributes
    ----------
    dataframe_backend:
        DataFrame backend
    laktory_root:
        Laktory cache root directory. Used when a pipeline needs to write
        checkpoint files.
    workspace_laktory_root:
        Root directory of a Databricks Workspace (excluding `"/Workspace") to
        which databricks objects like notebooks and workspace files are
        deployed.


    """

    dataframe_backend: str = None
    workspace_laktory_root: str = "/.laktory/"
    laktory_root: str = "/laktory/"

    @model_validator(mode="after")
    def apply_settings(self) -> Any:
        if self.dataframe_backend:
            settings.dataframe_backend = self.dataframe_backend

        if self.workspace_laktory_root:
            settings.workspace_laktory_root = self.workspace_laktory_root

        if self.laktory_root:
            settings.laktory_root = self.laktory_root

        return self


class Pulumi(BaseModel):
    """
    config:
        Pulumi configuration settings. Generally used to
        configure providers. See references for more details.
    outputs:
        Requested resources-related outputs. See references for details.

    References
    ----------
    - Pulumi [configuration](https://www.pulumi.com/docs/concepts/config/)
    - Pulumi [outputs](https://www.pulumi.com/docs/concepts/inputs-outputs/#outputs)
    """

    config: dict[str, str] = {}
    outputs: dict[str, str] = {}


class StackResources(BaseModel):
    """
    Resources definition for a given stack or stack environment.

    Attributes
    ----------
    databricks_alerts:
        Databricks Alerts
    databricks_dbfsfiles:
        Databricks DbfsFiles
    databricks_catalogs:
        Databricks Catalogs
    databricks_clusters:
        Databricks Clusters
    databricks_clusterpolicies:
        Databricks Cluster Policies
    databricks_dashboards:
        Databricks Dashboards
    databricks_directories:
        Databricks Directories
    databricks_dltpipelines:
        Databricks DLT Pipelines
    databricks_externallocations:
        Databricks External Locations
    databricks_groups:
        Databricks Groups
    databricks_grant:
        Databricks Grant
    databricks_grants:
        Databricks Grants
    databricks_jobs:
        Databricks Jobs
    databricks_metastoreassignments:
        Databricks Metastore Assignments
    databricks_metastoredataaccesses:
        Databricks Metastore Data Accesses
    databricks_metastores:
        Databricks Metastores
    databricks_mlflowexperiments:
        Databricks MLflow Experiments
    databricks_mlflowmodels:
        Databricks MLflow models
    databricks_mlflowwebhooks:
        Databricks MLflow webhooks
    databricks_networkconnectivityconfig
        Databricks Network Connectivity Config
    databricks_notebooks:
        Databricks Notebooks
    databricks_queries:
        Databricks Queries
    databricks_repo:
        Databricks Repo
    databricks_schemas:
        Databricks Schemas
    databricks_secretscopes:
        Databricks SecretScopes
    databricks_serviceprincipals:
        Databricks ServicePrincipals
    databricks_tables:
        Databricks Tables
    databricks_users:
        Databricks Users
    databricks_vectorsearchendpoint:
        Databricks Vector Search Endpoint
    databricks_vectorsearchindex:
        Databricks Vector Search Index
    databricks_volumes:
        Databricks Volumes
    databricks_warehouses:
        Databricks Warehouses
    databricks_workspacebindings:
        Databricks Workspace Bindings
    databricks_workspacefiles:
        Databricks WorkspacFiles
    pipelines:
        Laktory Pipelines
    providers:
        Providers
    """

    databricks_alerts: dict[str, Alert] = {}
    databricks_catalogs: dict[str, Catalog] = {}
    databricks_clusterpolicies: dict[str, ClusterPolicy] = {}
    databricks_clusters: dict[str, Cluster] = {}
    databricks_dashboards: dict[str, Dashboard] = {}
    databricks_dbfsfiles: dict[str, DbfsFile] = {}
    databricks_directories: dict[str, Directory] = {}
    databricks_dltpipelines: dict[str, DLTPipeline] = {}
    databricks_externallocations: dict[str, ExternalLocation] = {}
    databricks_grant: dict[str, Grant] = {}
    databricks_grants: dict[str, Grants] = {}
    databricks_groups: dict[str, Group] = {}
    databricks_jobs: dict[str, Job] = {}
    databricks_metastoreassignments: dict[str, MetastoreAssignment] = {}
    databricks_metastoredataaccesses: dict[str, MetastoreDataAccess] = {}
    databricks_metastores: dict[str, Metastore] = {}
    databricks_mlflowexperiments: dict[str, MLflowExperiment] = {}
    databricks_mlflowmodels: dict[str, MLflowModel] = {}
    databricks_mlflowwebhooks: dict[str, MLflowWebhook] = {}
    databricks_networkconnectivityconfig: dict[str, MwsNetworkConnectivityConfig] = {}
    databricks_notebooks: dict[str, Notebook] = {}
    databricks_queries: dict[str, Query] = {}
    databricks_repos: dict[str, Repo] = {}
    databricks_schemas: dict[str, Schema] = {}
    databricks_secrets: dict[str, Secret] = {}
    databricks_secretscopes: dict[str, SecretScope] = {}
    databricks_serviceprincipals: dict[str, ServicePrincipal] = {}
    databricks_tables: dict[str, Table] = {}
    databricks_users: dict[str, User] = {}
    databricks_vectorsearchendpoints: dict[str, VectorSearchEndpoint] = {}
    databricks_vectorsearchindexes: dict[str, VectorSearchIndex] = {}
    databricks_volumes: dict[str, Volume] = {}
    databricks_warehouses: dict[str, Warehouse] = {}
    databricks_workspacebindings: dict[str, WorkspaceBinding] = {}
    databricks_workspacefiles: dict[str, WorkspaceFile] = {}
    pipelines: dict[str, Pipeline] = {}
    providers: dict[
        str, Union[AWSProvider, AzureProvider, AzurePulumiProvider, DatabricksProvider]
    ] = {}

    @model_validator(mode="after")
    def update_resource_names(self) -> Any:
        for k, r in self._get_all().items():
            if r.resource_name_ and k != r.resource_name_:
                raise ValueError(
                    f"Provided resource name {r.resource_name_} does not match provided key {k}"
                )
            r.resource_name_ = k
        return self

    def _get_all(self, providers_excluded=False, providers_only=False):
        resources = {}
        for resource_type in self.model_fields.keys():
            if resource_type in ["variables"]:
                continue

            for resource_name, _r in getattr(self, resource_type).items():
                if resource_name in resources.keys():
                    raise ValueError(
                        f"Stack resource names are not unique. '{resource_name}' is already used."
                    )

                if providers_excluded and isinstance(_r, BaseProvider):
                    continue

                if providers_only and not isinstance(_r, BaseProvider):
                    continue

                resources[resource_name] = _r

        return resources


class EnvironmentStack(BaseModel):
    """
    Environment-specific stack definition.

    Attributes
    ----------
    backend:
        IaC backend used for deployment.
    description:
        Description of the stack
    name:
        Name of the stack. If Pulumi is used as a backend, it should match
        the name of the Pulumi project.
    organization:
        Organization
    pulumi:
        Pulumi-specific settings
    resources:
        Dictionary of resources to be deployed. Each key should be a resource
        type and each value should be a dictionary of resources who's keys are
        the resource names and the values the resources definitions.
    settings:
        Laktory settings
    terraform:
        Terraform-specific settings
    variables:
        Dictionary of variables made available in the resources definition.
    """

    backend: Literal["pulumi", "terraform"] = None
    description: str = None
    name: str
    organization: str = None
    pulumi: Pulumi = Pulumi()
    resources: Union[StackResources, None] = StackResources()
    settings: LaktorySettings = None
    terraform: Terraform = Terraform()
    variables: dict[str, Any] = {}


class EnvironmentSettings(BaseModel):
    """
    Settings overwrite for a specific environments

    Attributes
    ----------
    resources:
        Dictionary of resources to be deployed. Each key should be a resource
        type and each value should be a dictionary of resources who's keys are
        the resource names and the values the resources definitions.
    variables:
        Dictionary of variables made available in the resources definition.
    terraform:
        Terraform-specific settings
    """

    resources: Any = None
    variables: dict[str, Any] = None
    terraform: Terraform = Terraform()


class Stack(BaseModel):
    """
    The Stack defines a collection of deployable resources, the deployment
    configuration, some variables and the environment-specific settings.

    Attributes
    ----------
    backend:
        IaC backend used for deployment.
    description:
        Description of the stack
    environments:
        Environment-specific overwrite of config, resources or variables
        arguments.
    name:
        Name of the stack. If Pulumi is used as a backend, it should match
        the name of the Pulumi project.
    organization:
        Organization
    pulumi:
        Pulumi-specific settings
    resources:
        Dictionary of resources to be deployed. Each key should be a resource
        type and each value should be a dictionary of resources who's keys are
        the resource names and the values the resources definitions.
    settings:
        Laktory settings
    terraform:
        Terraform-specific settings
    variables:
        Dictionary of variables made available in the resources definition.

    Examples
    --------
    ```py
    from laktory import models

    stack = models.Stack(
        name="workspace",
        backend="pulumi",
        pulumi={
            "config": {
                "databricks:host": "${vars.DATABRICKS_HOST}",
                "databricks:token": "${vars.DATABRICKS_TOKEN}",
            },
        },
        resources={
            "databricks_dltpipelines": {
                "pl-stock-prices": {
                    "name": "pl-stock-prices",
                    "development": "${vars.is_dev}",
                    "libraries": [
                        {"notebook": {"path": "/pipelines/dlt_brz_template.py"}},
                    ],
                }
            },
            "databricks_jobs": {
                "job-stock-prices": {
                    "name": "job-stock-prices",
                    "clusters": [
                        {
                            "name": "main",
                            "spark_version": "14.0.x-scala2.12",
                            "node_type_id": "Standard_DS3_v2",
                        }
                    ],
                    "tasks": [
                        {
                            "task_key": "ingest",
                            "job_cluster_key": "main",
                            "notebook_task": {
                                "notebook_path": "/.laktory/jobs/ingest_stock_prices.py",
                            },
                        },
                        {
                            "task_key": "pipeline",
                            "depends_ons": [{"task_key": "ingest"}],
                            "pipeline_task": {
                                "pipeline_id": "${resources.dlt-pl-stock-prices.id}",
                            },
                        },
                    ],
                }
            },
        },
        variables={
            "org": "okube",
        },
        environments={
            "dev": {
                "variables": {
                    "is_dev": True,
                }
            },
            "prod": {
                "variables": {
                    "is_dev": False,
                }
            },
        },
    )
    ```

    """

    backend: Literal["pulumi", "terraform"] = None
    description: str = None
    environments: dict[str, EnvironmentSettings] = {}
    name: str
    organization: Union[str, None] = None
    pulumi: Pulumi = Pulumi()
    resources: Union[StackResources, None] = StackResources()
    settings: LaktorySettings = None
    terraform: Terraform = Terraform()
    variables: dict[str, Any] = {}
    _envs: dict[str, EnvironmentStack] = None

    @model_validator(mode="before")
    @classmethod
    def apply_settings(cls, data: Any) -> Any:
        """Required to apply settings before instantiating resources and setting default values"""
        settings = data.get("settings", None)
        if settings:
            if not isinstance(settings, dict):
                settings = settings.model_dump()
            LaktorySettings(**settings)

        return data

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def get_env(self, env_name: str) -> EnvironmentStack:
        """
        Complete definition the stack for a given environment. It takes into
        account both the default stack values and environment-specific
        overwrites.

        Parameters
        ----------
        env_name:
            Name of the environment

        Returns
        -------
        :
            Environment definitions.
        """

        if env_name is None:
            env = self
            env.push_vars()
            return env

        if env_name not in self.environments.keys():
            raise ValueError(f"Environment '{env_name}' is not declared in the stack.")

        if self._envs is None:
            ENV_FIELDS = ["pulumi", "resources", "terraform", "variables"]

            # Because some fields are excluded from the dump, they need to be
            # manually dumped and added back to the base dump
            def dump_with_excluded(obj: Any) -> Any:
                # Check data type, call recursively if not a BaseModel
                if isinstance(obj, list):
                    return [dump_with_excluded(v) for v in obj]
                elif isinstance(obj, dict):
                    return {k: dump_with_excluded(v) for k, v in obj.items()}
                elif not isinstance(obj, BaseModel):
                    return obj

                # Get model dump
                model = obj
                data = model.model_dump(exclude_unset=True)

                # Loop through all model fields
                for field_name, field in model.model_fields.items():
                    # Explicitly dump excluded fields - variables
                    if field_name == "variables" and model.variables is not None:
                        data["variables"] = copy.deepcopy(model.variables)

                    # Explicitly dump excluded fields - resource options
                    if field_name == "options" and field.annotation == ResourceOptions:
                        data["options"] = model.options.model_dump(exclude_unset=True)

                    # Explicitly dump excluded fields - resource name
                    if field_name == "resource_name_" and model.resource_name_:
                        data["resource_name_"] = model.resource_name_

                    # Explicitly dump excluded fields - lookup existing
                    if field_name == "lookup_existing" and model.lookup_existing:
                        data["lookup_existing"] = model.lookup_existing.model_dump(
                            exclude_unset=True
                        )

                    # Parse list
                    if isinstance(data.get(field_name, None), list):
                        data[field_name] = [
                            dump_with_excluded(v) for v in getattr(model, field_name)
                        ]

                    # Parse dict (might result from a dict or a BaseModel)
                    elif isinstance(data.get(field_name, None), dict):
                        a = getattr(model, field_name)

                        if isinstance(a, dict):
                            for k in a.keys():
                                data[field_name][k] = dump_with_excluded(a[k])
                        else:
                            data[field_name] = dump_with_excluded(a)

                return data

            envs = {}
            for _env_name, env in self.environments.items():
                d = dump_with_excluded(self)
                _envs = d.pop("environments")

                for k in ENV_FIELDS:
                    v1 = _envs[_env_name].get(k, {})
                    if k in d:
                        d[k] = merge_dicts(d[k], v1)
                    elif k in _envs[_env_name]:
                        d[k] = v1

                envs[_env_name] = EnvironmentStack(**d)
                envs[_env_name].push_vars()

            self._envs = envs

        return self._envs[env_name]

    # ----------------------------------------------------------------------- #
    # Pulumi Methods                                                          #
    # ----------------------------------------------------------------------- #

    def to_pulumi(self, env_name: Union[str, None] = None):
        """
        Create a pulumi stack for a given environment `env`.

        Parameters
        ----------
        env_name:
            Target environment. If `None`, used default stack values only.

        Returns
        -------
        : PulumiStack
            Pulumi-specific stack definition
        """
        from laktory.models.stacks.pulumistack import PulumiStack

        env = self.get_env(env_name=env_name).inject_vars()

        # Resources
        resources = {}
        for r in env.resources._get_all().values():
            for _r in r.core_resources:
                resources[_r.resource_name] = _r

        return PulumiStack(
            name=env.name,
            organization=env.organization,
            config=env.pulumi.config,
            description=env.description,
            resources=resources,
            outputs=env.pulumi.outputs,
        )

    # ----------------------------------------------------------------------- #
    # Terraform Methods                                                       #
    # ----------------------------------------------------------------------- #

    def to_terraform(self, env_name: Union[str, None] = None):
        """
        Create a terraform stack for a given environment `env`.

        Parameters
        ----------
        env_name:
            Target environment. If `None`, used default stack values only.

        Returns
        -------
        : TerraformStack
            Terraform-specific stack definition
        """
        from laktory.models.stacks.terraformstack import TerraformStack

        env = self.get_env(env_name=env_name).inject_vars()

        # Providers
        providers = {}
        for r in env.resources._get_all(providers_only=True).values():
            for _r in r.core_resources:
                rname = _r.resource_name
                providers[rname] = _r

        # Resources
        resources = {}
        for r in env.resources._get_all(providers_excluded=True).values():
            for _r in r.core_resources:
                resources[_r.resource_name] = _r

        # Update terraform
        return TerraformStack(
            terraform={"backend": env.terraform.backend},
            providers=providers,
            resources=resources,
        )



import json
from typing import Any
from typing import Union

from pydantic import field_validator

from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class ClusterPolicyLibraryCran(BaseModel):
    package: str
    repo: str = None


class ClusterPolicyLibraryMaven(BaseModel):
    coordinates: str
    exclusions: list[str] = None
    repo: str = None


class ClusterPolicyLibraryPypi(BaseModel):
    package: str = None
    repo: str = None


class ClusterPolicyLibrary(BaseModel):
    """
    Cluster Policy Library

    Attributes
    ----------
    cran:
        Cran library specifications
    egg:
        Egg filepath
    jar:
        Jar filepath
    maven:
        TODO
    pypi:
        Pypi library specifications
    requirements:
        TODO
    whl:
        Wheel filepath
    """

    cran: ClusterPolicyLibraryCran = None
    egg: str = None
    jar: str = None
    maven: ClusterPolicyLibraryMaven = None
    pypi: ClusterPolicyLibraryPypi = None
    requirement: str = None
    whl: str = None


# class ClusterLookup(ResourceLookup):
#     """
#     Attributes
#     ----------
#     cluster_id:
#         The id of the cluster
#     """
#
#     cluster_id: str = Field(serialization_alias="id")


class ClusterPolicy(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks cluster policy

    Attributes
    ----------

    access_controls:
        List of access controls
    definition:
        Policy definition: JSON document expressed in [Databricks Policy
        Definition Language](https://docs.databricks.com/en/admin/clusters/policies.html#cluster-policy-definition).
        Cannot be used with `policy_family_id`.
    description:
        Additional human-readable description of the cluster policy.
    libraries:
        TODO
    max_clusters_per_user:
        Maximum number of clusters allowed per user. When omitted, there is no
        limit. If specified, value must be greater than zero.
    name:
        Cluster policy name. This must be unique. Length must be between 1 and
        100 characters.
    policy_family_definition_overrides:
        Policy definition JSON document expressed in Databricks Policy
        Definition Language. The JSON document must be passed as a string and
         cannot be embedded in the requests. You can use this to customize the
         policy definition inherited from the policy family. Policy rules
         specified here are merged into the inherited policy definition.
    policy_family_id:
        ID of the policy family. The cluster policy's policy definition
        inherits the policy family's policy definition. Cannot be used with
        `definition`. Use `policy_family_definition_overrides` instead to
        customize the policy definition.

    Examples
    --------
    ```py
    from laktory import models

    cluster = models.resources.databricks.ClusterPolicy(
        name="okube",
        definition={
            "dbus_per_hour": {
                "type": "range",
                "maxValue": 10,
            },
            "autotermination_minutes": {"type": "fixed", "value": 30, "hidden": True},
            "custom_tags.team": {
                "type": "fixed",
                "value": "okube",
            },
        },
        libraries=[
            {
                "pypi": {
                    "package": "laktory==0.5.0",
                }
            }
        ],
        access_controls=[
            {"permission_level": "CAN_USE", "group_name": "account users"}
        ],
    )
    ```

    References
    ----------

    * [Databricks Cluster Policy](https://docs.databricks.com/api/workspace/clusterpolicies/create)
    * [Pulumi Databricks Cluster Policy](https://www.pulumi.com/registry/packages/databricks/api-docs/clusterpolicy)

    """

    access_controls: list[AccessControl] = []
    definition: Union[str, dict[str, Any]] = None
    description: str = None
    libraries: list[ClusterPolicyLibrary] = None
    max_clusters_per_user: int = None
    name: str
    policy_family_definition_overrides: str = None
    policy_family_id: str = None

    @field_validator("definition")
    def validate_type(cls, v: Union[str, dict[str, str]]) -> str:
        if isinstance(v, dict):
            v = json.dumps(v)
        return v

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    cluster_policy_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:ClusterPolicy"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "libraries": "libraries",
        }

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_cluster_policy"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import warnings
from typing import Any
from typing import Literal
from typing import Union

from pydantic import model_validator

from laktory._logger import get_logger
from laktory.exceptions import DataQualityCheckFailedError
from laktory.models.basemodel import BaseModel
from laktory.models.dataframecolumnexpression import DataFrameColumnExpression
from laktory.models.dataquality.check import DataQualityCheck
from laktory.typing import AnyDataFrame
from laktory.typing import AnyDataFrameColumn

logger = get_logger(__name__)


class ExpectationTolerance(BaseModel):
    """
    Tolerance values for data quality expectations with support for either
    absolute or relative tolerances.

    Attributes
    ----------
    abs:
        Maximum number of rows with failure for a PASS status
    rel:
        Relative number of rows with failure for a PASS status
    """

    abs: int = None
    rel: float = None

    @model_validator(mode="after")
    def at_least_one(self) -> Any:
        if self.abs is None and self.rel is None:
            raise ValueError("At least `abs` or `rel` must be set.")

        if not (self.abs is None or self.rel is None):
            raise ValueError("Only one of `abs` or `rel` must be set.")

        return self


class DataQualityExpectation(BaseModel):
    """
    Data Quality Expectation for a given DataFrame expressed as a row-specific
    condition (`type="ROW"`) or as an aggregated metric (`type="AGGREGATE"`).

    The expression may be defined as a SQL statement or a DataFrame expression.

    Attributes
    ----------
    action:
        Action to take when expectation is not met.
        `WARN`: Write invalid records to the output DataFrame, but log
        exception.
        `DROP`: Drop Invalid records to the output DataFrame and log exception.
        `QUARANTINE`: Forward invalid data for quarantine.
        `FAIL`: Raise exception when invalid records are found.
    type:
        Type of expectation:
        `"ROW"`: Row-specific condition. Must be a boolean expression.
        `"AGGREGATE"`: Global condition. Must be a boolean expression.
    name:
        Name of the expectation
    expr:
        SQL or DataFrame expression representing a row-specific condition or
        an aggregated metric.
    tolerance:
        Tolerance for non-matching rows before resulting in failure. Only
        available for "ROW" type expectation.

    Examples
    --------
    ```py
    from laktory import models

    dqe = models.DataQualityExpectation(
        name="price higher than 10",
        action="WARN",
        expr="close > 127",
        tolerance={"rel": 0.05},
    )
    print(dqe)
    '''
    variables={} action='WARN' type='ROW' name='price higher than 10' expr=DataFrameColumnExpression(variables={}, value='close > 127', type='SQL') tolerance=ExpectationTolerance(variables={}, abs=None, rel=0.05)
    '''

    dqe = models.DataQualityExpectation(
        name="rows count",
        expr="COUNT(*) > 50",
        type="AGGREGATE",
    )
    print(dqe)
    '''
    variables={} action='WARN' type='AGGREGATE' name='rows count' expr=DataFrameColumnExpression(variables={}, value='COUNT(*) > 50', type='SQL') tolerance=ExpectationTolerance(variables={}, abs=0, rel=None)
    '''
    ```

    References
    ----------

    * [DLT Table Expectations](https://docs.databricks.com/en/delta-live-tables/expectations.html)
    """

    action: Literal["WARN", "DROP", "QUARANTINE", "FAIL"] = "WARN"
    type: Literal["AGGREGATE", "ROW"] = "ROW"
    name: str
    expr: Union[str, DataFrameColumnExpression] = None
    tolerance: ExpectationTolerance = ExpectationTolerance(abs=0)
    _dataframe_backend: Literal["SPARK", "POLARS"] = None
    _check: DataQualityCheck = None

    @model_validator(mode="after")
    def parse_expr(self) -> Any:
        if isinstance(self.expr, str):
            self.expr = DataFrameColumnExpression(value=self.expr)
        return self

    @model_validator(mode="after")
    def validate_action(self) -> Any:
        if self.type == "AGGREGATE" and self.action in ["DROP", "QUARANTINE"]:
            raise ValueError(
                f"'{self.type}' action is not supported for 'AGGREGATE' type."
            )
        return self

    @model_validator(mode="after")
    def warn_invalid_type(self):
        msg = self.type_warning_msg
        if msg:
            import warnings

            warnings.warn(msg)
        return self

    # ----------------------------------------------------------------------- #
    # Filters                                                                 #
    # ----------------------------------------------------------------------- #

    @property
    def pass_filter(self) -> Union[AnyDataFrameColumn, None]:
        """Expression representing all rows meeting the expectation."""
        return self.expr.eval(dataframe_backend=self._dataframe_backend)

    @property
    def fail_filter(self) -> Union[AnyDataFrameColumn, None]:
        """Expression representing all rows not meeting the expectation."""
        return ~self.expr.eval(dataframe_backend=self._dataframe_backend)

    @property
    def keep_filter(self) -> Union[AnyDataFrameColumn, None]:
        """
        Expression representing all rows to keep, considering both the
        expectation and the selected action.
        """
        # if self._check is None:
        #     raise ValueError()
        if self.type != "ROW":
            return None
        if self.action not in ["DROP", "QUARANTINE"]:
            return None
        # if self._check.fails_count == 0:
        #     return None
        return self.pass_filter

    @property
    def quarantine_filter(self) -> Union[AnyDataFrameColumn, None]:
        """
        Expression representing all rows to quarantine, considering both the
        expectation and the selected action.
        """
        # if self._check is None:
        #     raise ValueError()
        if self.type != "ROW":
            return None
        if self.action not in ["QUARANTINE"]:
            return None
        # if self._check.fails_count == 0:
        #     return None
        return self.fail_filter

    # ----------------------------------------------------------------------- #
    # Compatibility                                                           #
    # ----------------------------------------------------------------------- #

    @property
    def is_dlt_compatible(self):
        return self.expr.type == "SQL" and self.type == "ROW"

    @property
    def is_streaming_compatible(self):
        return (
            self.type == "ROW"
            and self.tolerance.rel is None
            and self.tolerance.abs == 0
        )

    # ----------------------------------------------------------------------- #
    # Logging                                                                 #
    # ----------------------------------------------------------------------- #

    @property
    def type_warning_msg(self):
        msg = None
        if self.type != "AGGREGATE":
            for k in [
                "count(",
                "sum(",
                "avg(",
                "mean(",
                "max(",
                "min(",
                "variance(",
                "stddev(",
                "row_number(",
                "rank(",
            ]:
                if k in self.expr.value.lower():
                    msg = f"'ROW' type is selected for expectation '{self.name}' ({self.expr.value}). Should probably be 'AGGREGATE'."
                    break
        return msg

    @property
    def log_msg(self) -> str:
        msg = f"expr: {self.expr.value} | status: {self.check.status}"
        if self.type == "ROW" and self.check.fails_count:
            msg += f" | fails count: {self.check.fails_count} / {self.check.rows_count} ({100 * self.check.failure_rate:5.2f} %)"
        return msg

    # ----------------------------------------------------------------------- #
    # Check                                                                   #
    # ----------------------------------------------------------------------- #

    @property
    def check(self):
        return self._check

    def run_check(
        self,
        df: AnyDataFrame,
        raise_or_warn: bool = False,
        node=None,
    ) -> DataQualityCheck:
        """
        Check if expectation is met save result.

        Parameters
        ----------
        df:
            Input DataFrame for checking the expectation.
        raise_or_warn:
            Raise exception or issue warning if expectation is not met.
        node:
            Pipeline Node

        Returns
        -------
        output: DataQualityCheck
            Check result.
        """

        logger.info(
            f"Checking expectation '{self.name}' | {self.expr.value} (type: {self.type})"
        )

        # Assign DataFrame type
        dtype = str(type(df)).lower()
        if "spark" in dtype:
            self._dataframe_backend = "SPARK"
        elif "polars" in dtype:
            self._dataframe_backend = "POLARS"
        else:
            raise ValueError(f"DataFrame type '{dtype}' not supported")

        # Run Check
        self._check = self._check_df(df)

        if raise_or_warn:
            self.raise_or_warn(node)

        return self._check

    def _check_df(self, df):
        if self._dataframe_backend == "SPARK":
            rows_count = df.count()
        elif self._dataframe_backend == "POLARS":
            import polars as pl

            rows_count = df.select(pl.len()).collect().item()

        if rows_count == 0:
            _check = DataQualityCheck(
                fails_count=0,
                status="PASS",
                rows_count=0,
            )
            return _check

        if self.type == "ROW":
            try:
                df_fail = df.filter(self.fail_filter)
            except Exception as e:
                if "Rewrite the query to avoid window functions" in getattr(
                    e, "desc", ""
                ):
                    e.desc += f"\n{self.type_warning_msg}"
                raise e

            if self._dataframe_backend == "SPARK":
                fails_count = df_fail.count()
            elif self._dataframe_backend == "POLARS":
                import polars as pl

                fails_count = df_fail.select(pl.len()).collect().item()

            status = "PASS"
            if self.tolerance.abs is not None:
                if fails_count > self.tolerance.abs:
                    status = "FAIL"
            elif self.tolerance.rel is not None:
                if rows_count > 0 and fails_count / rows_count > self.tolerance.rel:
                    status = "FAIL"

            _check = DataQualityCheck(
                fails_count=fails_count,
                status=status,
                rows_count=rows_count,
            )
            failure_str = f"({100 * _check.failure_rate:5.2f}%)"
            if status == "PASS":
                logger.info(f"Checking expectation '{self.name}' | status : {status}")
            else:
                logger.info(
                    f"Checking expectation '{self.name}' | status : {status} - failed rows : {fails_count} {failure_str}"
                )
            return _check

        if self.type == "AGGREGATE":
            import pyspark.sql.functions as F  # noqa: F401

            if self.expr.type == "SQL":
                _df = df.select(self.expr.eval()).toPandas()
            else:
                _df = df.agg(self.expr.eval()).toPandas()

            status = _df.iloc[0].values[0]

            status = "PASS" if status else "FAIL"

            _check = DataQualityCheck(
                status=status,
                rows_count=rows_count,
            )
            logger.info(f"Checking expectation '{self.name}' | status : {status}")
            return _check

    def raise_or_warn(self, node=None) -> None:
        """
        Raise exception or issue warning if expectation is not met.
        """

        # Failure Message
        msg = f"Expectation '{self.name}'"
        if node:
            msg += f" for node '{node.name}'"
        msg += f" FAILED | {self.log_msg}"

        if self.check.status != "FAIL":
            return

        # Raise Exception
        if self.action == "FAIL":
            raise DataQualityCheckFailedError(self, node)
        else:
            # actions: WARN, DROP, QUARANTINE
            warnings.warn(msg)



import polars as pl

__all__ = [
    "string_split",
    "uuid",
]


# --------------------------------------------------------------------------- #
# string_split                                                                #
# --------------------------------------------------------------------------- #


def string_split(
    x: pl.Expr,
    pattern: str,
    key: int,
) -> pl.Expr:
    """
    Get substring using separator `pat`.

    Parameters
    ----------
    x:
        Input text series to split
    pattern:
        String or regular expression to split on. If not specified, split on whitespace.
    key:
        Split index to return

    Returns
    -------
    :
        Result

    Examples
    --------
    ```py
    import polars as pl

    import laktory  # noqa: F401

    df = pl.DataFrame({"x": ["price_close"]})

    df = df.with_columns(
        y=pl.Expr.laktory.string_split(pl.col("x"), pattern="_", key=1)
    )
    print(df.glimpse(return_as_string=True))
    '''
    Rows: 1
    Columns: 2
    $ x <str> 'price_close'
    $ y <str> 'close'
    '''
    ```
    """
    return x.str.split(by=pattern).list.get(key, null_on_oob=True)


# --------------------------------------------------------------------------- #
# uuid                                                                        #
# --------------------------------------------------------------------------- #


def uuid() -> pl.Expr:
    """
    Create a unique id for each row.

    Returns
    -------
    :
        Output column

    Examples
    --------
    ```py
    import polars as pl

    import laktory  # noqa: F401

    df = pl.DataFrame({"id": [0, 1, 2]})
    df = df.with_columns(uuid=pl.Expr.laktory.uuid())
    ```
    """

    def generate_uuid():
        import uuid

        return str(uuid.uuid4())

    return pl.first().map_elements(lambda _: generate_uuid(), return_dtype=pl.Utf8)



from typing import Any
from typing import Literal
from typing import Union

from pydantic import model_validator

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class VectorSearchIndexDeltaSyncIndexSpecEmbeddingSourceColumn(BaseModel):
    """
    Attributes
    ----------
    embedding_model_endpoint_name:
        The name of the embedding model endpoint
    name:
        Three-level name of the Vector Search Index to create
        (catalog.schema.index_name).
    """

    embedding_model_endpoint_name: str
    name: str


class VectorSearchIndexDeltaSyncIndexSpecEmbeddingVectorColumn(BaseModel):
    """
    Attributes
    ----------
    embedding_dimension:
        Dimension of the embedding vector.
    name:
        Three-level name of the Vector Search Index to create
        (catalog.schema.index_name).
    """

    embedding_dimension: int
    name: str


class VectorSearchIndexDeltaSyncIndexSpec(BaseModel):
    """
    Attributes
    ----------
    embedding_source_columns:
        Array of objects representing columns that contain the embedding source
    embedding_vector_columns:
        (required if embedding_source_columns isn't provided) array of objects
        representing columns that contain the embedding vectors.
    embedding_writeback_table:
        TODO
    pipeline_id:
        ID of the associated Delta Live Table pipeline.
    pipeline_type:
        Pipeline execution mode. Possible values are:
            TRIGGERED: If the pipeline uses the triggered execution mode, the
            system stops processing after successfully refreshing the source
            table in the pipeline once, ensuring the table is updated based on
            the data available when the update started.
            CONTINUOUS: If the pipeline uses continuous execution, the pipeline
            processes new data as it arrives in the source table to keep the
            vector index fresh.
    """

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "embedding_source_columns": "embedding_source_columns",
            "embedding_vector_columns": "embedding_vector_columns",
        }

    embedding_source_columns: list[
        VectorSearchIndexDeltaSyncIndexSpecEmbeddingSourceColumn
    ] = None
    embedding_vector_columns: list[
        VectorSearchIndexDeltaSyncIndexSpecEmbeddingVectorColumn
    ] = None
    embedding_writeback_table: str = None
    pipeline_id: str = None
    pipeline_type: Literal["TRIGGERED", "CONTINUOUS"] = None
    source_table: str = None


class VectorSearchIndexDirectAccessIndexSpec(BaseModel):
    """
    Attributes
    ----------
    embedding_source_columns:
        Array of objects representing columns that contain the embedding source
    embedding_vector_columns:
        (required if embedding_source_columns isn't provided) array of objects
        representing columns that contain the embedding vectors.
    schema_json:
        The schema of the index in JSON format. Check the API documentation for
        a list of supported data types.
    """

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "embedding_source_columns": "embedding_source_columns",
            "embedding_vector_columns": "embedding_vector_columns",
        }

    embedding_source_columns: list[
        VectorSearchIndexDeltaSyncIndexSpecEmbeddingSourceColumn
    ] = None
    embedding_vector_columns: list[
        VectorSearchIndexDeltaSyncIndexSpecEmbeddingVectorColumn
    ] = None
    # schema_json: str


class VectorSearchIndex(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Warehouse

    Attributes
    ----------
    access_controls:
        Vector search endpoint access controls
    endpoint_name:
        The name of the Vector Search Endpoint that will be used for indexing
        the data.
    index_type:
        Vector Search index type. Currently supported values are:
        DELTA_SYNC: An index that automatically syncs with a source Delta
                    Table, automatically and incrementally updating the index
                    as the underlying data in the Delta Table changes.
        DIRECT_ACCESS: An index that supports the direct read and write of
                       vectors and metadata through our REST and SDK APIs.
                       With this model, the user manages index updates.
    primary_key:
        The column name that will be used as a primary key.
    delta_sync_index_spec:
        Specification for Delta Sync Index. Required if index_type is
        DELTA_SYNC.
    direct_access_index_spec:
        Specification for Direct Vector Access Index. Required if index_type is
         DIRECT_ACCESS.
    name:
        Three-level name of the Vector Search Index to create (catalog.schema.index_name).


    Examples
    --------
    ```py
    from laktory import models

    index = models.resources.databricks.VectorSearchIndex(
        name="dev.finances.market_news_vs_index",
        primary_key="id",
        endpoint_name="default",
        index_type="DELTA_SYNC",
        delta_sync_index_spec={
            "source_table": "dev.finances.market_news",
            "embedding_vector_columns": [
                {"name": "embedding", "embedding_dimension": 4095}
            ],
            "pipeline_type": "TRIGGERED",
        },
    )
    ```
    """

    # access_controls: list[AccessControl] = []
    endpoint_name: str
    index_type: Literal["DELTA_SYNC", "DIRECT_ACCESS"]
    primary_key: str
    delta_sync_index_spec: VectorSearchIndexDeltaSyncIndexSpec = None
    direct_access_index_spec: VectorSearchIndexDirectAccessIndexSpec = None
    name: str

    @model_validator(mode="after")
    def check_index_spec(self) -> Any:
        if self.index_type == "DELTA_SYNC" and self.delta_sync_index_spec is None:
            raise ValueError(
                "`delta_sync_index_spec` must be set with `index_type` = 'DELTA_SYNC'"
            )

        if self.index_type == "DIRECT_ACCESS" and self.direct_access_index_spec is None:
            raise ValueError(
                "`direct_access_index_spec` must be set with `index_type` = 'DIRECT_ACCESS'"
            )

        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:VectorSearchIndex"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_vector_search_index"

    # @property
    # def terraform_resource_lookup_type(self) -> str:
    #     return "databricks_sql_warehouse"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



class SparkFactory:
    def __init__(self):
        self._spark = None

    @property
    def spark(self):
        if self._spark is None:
            from pyspark.sql import SparkSession

            spark = (
                SparkSession.builder.appName("UnitTesting")
                .config("spark.jars.packages", "io.delta:delta-spark_2.12:3.2.0")
                .config(
                    "spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension"
                )
                .config(
                    "spark.sql.catalog.spark_catalog",
                    "org.apache.spark.sql.delta.catalog.DeltaCatalog",
                )
                .getOrCreate()
            )
            spark.conf.set("spark.sql.session.timeZone", "UTC")
            self._spark = spark
        return self._spark



from typing import TYPE_CHECKING
from typing import Union

from pydantic import BaseModel
from pyspark.sql.dataframe import DataFrame

from laktory._logger import get_logger

if TYPE_CHECKING:
    from laktory.models.transformers.basechainnode import ChainNodeColumn


logger = get_logger(__name__)


class TimeWindow(BaseModel):
    """
    Specifications for Time Window Aggregation

    Attributes
    ----------
    time_column:
        Timestamp column used for grouping rows
    window_duration:
        Duration of the window e.g. ‘1 second’, ‘1 day 12 hours’, ‘2 minutes’
    slide_duration
        Duration of the slide. If a slide is smaller than the window, windows
        are overlapping
    start_time
        Offset with respect to 1970-01-01 00:00:00 UTC with which to start
        window intervals.

    References
    ----------

    * [spark structured streaming windows](https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html#types-of-time-windows)
    * [pyspark window](https://spark.apache.org/docs/3.1.3/api/python/reference/api/pyspark.sql.functions.window.html)

    """

    time_column: str
    window_duration: str
    slide_duration: Union[str, None] = None
    start_time: Union[str, None] = None


def groupby_and_agg(
    df,
    groupby_window: TimeWindow = None,
    groupby_columns: list[str] = None,
    agg_expressions: list["ChainNodeColumn"] = None,
) -> DataFrame:
    """
    Apply a groupby and create aggregation columns.

    Parameters
    ----------
    df:
        DataFrame
    groupby_window:
        Aggregation window definition
    groupby_columns:
        List of column names to group by
    agg_expressions:
        List of columns defining the aggregations

    Examples
    --------
    ```py
    import pandas as pd

    import laktory  # noqa: F401

    df0 = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["AAPL", "GOOGL"],
                "price": [200.0, 205.0],
                "tstamp": ["2023-09-01", "2023-09-01"],
            }
        )
    )

    df = df0.laktory.groupby_and_agg(
        groupby_window={
            "time_column": "tstamp",
            "window_duration": "1 day",
        },
        agg_expressions=[
            {
                "name": "mean_price",
                "expr": "F.mean('price')",
            },
        ],
    )

    print(df.toPandas().to_string())
    '''
                                           window  mean_price
    0  (2023-09-01 00:00:00, 2023-09-02 00:00:00)       202.5
    '''
    ```

    References
    ----------

    * [pyspark window](https://spark.apache.org/docs/3.1.3/api/python/reference/api/pyspark.sql.functions.window.html)
    """
    import pyspark.sql.functions as F

    from laktory.models.transformers.basechainnode import ChainNodeColumn

    # Parse inputs
    if groupby_window and not isinstance(groupby_window, TimeWindow):
        groupby_window = TimeWindow(**groupby_window)
    if agg_expressions is None:
        raise ValueError("`agg_expressions` must be specified")
    if groupby_columns is None:
        groupby_columns = []

    logger.info(
        f"Executing groupby ({groupby_window} & {groupby_columns}) with {agg_expressions}"
    )

    # Groupby arguments
    groupby = []
    if groupby_window:
        groupby += [
            F.window(
                timeColumn=groupby_window.time_column,
                windowDuration=groupby_window.window_duration,
                slideDuration=groupby_window.slide_duration,
                startTime=groupby_window.start_time,
            )
        ]

    for c in groupby_columns:
        groupby += [c]

    # Agg arguments
    aggs = []
    for expr in agg_expressions:
        if not isinstance(expr, ChainNodeColumn):
            expr = ChainNodeColumn(**expr)

        expr.type = None
        aggs += [expr.eval(dataframe_backend="SPARK").alias(expr.name)]

    return df.groupby(groupby).agg(*aggs)



from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.mwspermissionassignment import (
    MwsPermissionAssignment,
)
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class GroupLookup(ResourceLookup):
    """
    Attributes
    ----------
    id:
        Id of the group. Only supported when using Pulumi backend.
    display_name:
        Display name of the group. Only support when using Terraform backend
    """

    id: str = None
    display_name: str = None


class Group(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks group

    Attributes
    ----------
    acl_principal_id:
        Identifier for use in databricks_access_control_rule_set, e.g. `groups/Some Group`.
    allow_cluster_create:
        This is a field to allow the group to have cluster create privileges. More
        fine grained permissions could be assigned with databricks.Permissions and
        cluster_id argument. Everyone without `allow_cluster_create` argument set, but
        with permission to use Cluster Policy would be able to create clusters, but
        within boundaries of that specific policy.
    allow_instance_pool_create:
        This is a field to allow the group to have instance pool create privileges. More
        fine grained permissions could be assigned with databricks.Permissions and
        instance_pool_id argument.
    databricks_sql_access:
        This is a field to allow the group to have access to Databricks SQL feature in
        User Interface and through databricks_sql_endpoint.
    display_name:
        Display name for the group.
    external_id:
        ID of the group in an external identity provider.
    force:
        Ignore `cannot create group: Group with name X already exists.` errors and
        implicitly import the specific group into IaC state, enforcing entitlements
        defined in the instance of resource. This functionality is experimental and is
        designed to simplify corner cases, like Azure Active Directory synchronisation.
    lookup_existing:
        Specifications for looking up existing resource. Other attributes will
        be ignored.
    workspace_access
        When `True`, the group is allowed to have workspace access

    Examples
    --------
    ```py
    from laktory import models

    d = models.resources.databricks.Group(display_name="role-engineers")
    ```
    """

    acl_principal_id: str = None
    allow_cluster_create: bool = False
    allow_instance_pool_create: bool = None
    databricks_sql_access: bool = None
    display_name: str = None
    external_id: str = None
    force: bool = None
    lookup_existing: GroupLookup = Field(None, exclude=True)
    url: str = None
    workspace_access: bool = None
    workspace_permission_assignments: list[MwsPermissionAssignment] = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """display name"""
        return self.display_name

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Group"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["workspace_permission_assignments"]

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - workspace permission assignments
        """
        resources = []
        if self.workspace_permission_assignments:
            for a in self.workspace_permission_assignments:
                if a.principal_id is None:
                    a.principal_id = f"${{resources.{self.resource_name}.id}}"
                resources += [a]
        return resources

    # TODO:
    # if group.id is None:
    #     self.group = databricks.Group(name, opts=opts, **group.model_pulumi_dump())
    #     group.id = self.group.id
    # else:
    #     self.group = databricks.Group.get(name, id=group.id)

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_group"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import pyspark.sql.functions as F
from pyspark.sql.column import Column

from laktory.spark.functions._common import COLUMN_OR_NAME
from laktory.spark.functions._common import FLOAT_OR_COLUMN
from laktory.spark.functions._common import _col
from laktory.spark.functions._common import _lit

__all__ = [
    "roundp",
]


# --------------------------------------------------------------------------- #
# Rounding                                                                    #
# --------------------------------------------------------------------------- #


def roundp(
    x: COLUMN_OR_NAME,
    p: FLOAT_OR_COLUMN = 1.0,
) -> Column:
    """
    Evenly round to the given precision

    Parameters
    ------
    x:
        Input column
    p:
        Precision

    Returns
    -------
    :
        Output column

    Examples
    --------
    ```py
    import pyspark.sql.functions as F

    import laktory  # noqa: F401

    df = spark.createDataFrame([[0.781], [13.0]], ["x"])
    df = df.withColumn("y", F.laktory.roundp("x", p=5))
    print(df.laktory.show_string())
    '''
    +-----+----+
    |    x|   y|
    +-----+----+
    |0.781| 0.0|
    | 13.0|15.0|
    +-----+----+
    '''

    df = df.withColumn("y", F.laktory.roundp("x", p=0.25))
    print(df.laktory.show_string())
    '''
    +-----+----+
    |    x|   y|
    +-----+----+
    |0.781|0.75|
    | 13.0|13.0|
    +-----+----+
    '''
    ```
    """
    # eps0 = 1.0e-16
    # precision = float(precision)
    # if precision < eps0:
    #     raise ValueError("Precision must be greater than 1.0e-16")
    return F.round(_col(x) / _lit(p)) * _lit(p)



QUICKSTART_TEMPLATES = [
    "unity-catalog",
    "workspace",
    "workflows",
    "local-pipeline",
]

SUPPORTED_BACKENDS = [
    "pulumi",
    "terraform",
]

SUPPORTED_DFTYPES = [
    "POLARS",
    "SPARK",
]

SUPPORTED_DATATYPES = [
    "binary",
    "byte",
    "int8",
    "tinyint",
    "short",
    "int16",
    "smallint",
    "int",
    "int32",
    "long",
    "int64",
    "bigint",
    "float",
    "float32",
    "double",
    "float64",
    "boolean",
    "string",
    "utf8",
    "date",
    "timestamp",
    "datetime",
    # "struct<>",
    # "struct",
    # "array",
    # "list",
]

# CACHE_ROOT = "./laktory"
CACHE_ROOT = "./"

# for k in vars(types):
#     o = getattr(types, k)
#     if "Type" in k and inspect.isclass(o):
#         try:
#             typ = o()
#         except TypeError:
#             continue
#
#         if not hasattr(typ, "simpleString"):
#             continue
#
#         SUPPORTED_DATATYPES[typ.simpleString()] = typ



from typing import Literal
from typing import Union

from laktory.models.basemodel import BaseModel
from laktory.models.grants.volumegrant import VolumeGrant
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class Volume(BaseModel, PulumiResource, TerraformResource):
    """
    Volumes are Unity Catalog objects representing a logical volume of storage
    in a cloud object storage location. Volumes provide capabilities for
    accessing, storing, governing, and organizing files. While tables provide
    governance over tabular datasets, volumes add governance over non-tabular
    datasets. You can use volumes to store and access files in any format,
    including structured, semi-structured, and unstructured data.

    Attributes
    ----------
    name:
        Name of the volume
    catalog_name:
        Name of the catalog storing the volume
    grant:
        Grant(s) operating on the Volume and authoritative for a specific principal.
        Other principals within the grants are preserved. Mutually exclusive with
        `grants`.
    grants:
        Grants operating on the Volume and authoritative for all principals.
        Replaces any existing grants defined inside or outside of Laktory. Mutually
        exclusive with `grant`.
    schema_name:
        Name of the schema storing the volume
    storage_location:
        Path inside an External Location. Only used for EXTERNAL Volumes.
    volume_type:
        Type of volume. A managed volume is a Unity Catalog-governed storage volume created within the default storage
        location of the containing schema. An external volume is a Unity Catalog-governed storage volume registered
        against a directory within an external location.

    Examples
    --------
    ```py
    from laktory import models

    volume = models.resources.databricks.Volume(
        name="landing",
        catalog_name="dev",
        schema_name="sources",
        volume_type="EXTERNAL",
        storage_location="abfss://landing@lakehouse-storage.dfs.core.windows.net/",
        grants=[
            {"principal": "account users", "privileges": ["READ_VOLUME"]},
            {"principal": "role-metastore-admins", "privileges": ["WRITE_VOLUME"]},
        ],
    )
    print(volume.full_name)
    # > dev.sources.landing
    print(volume.parent_full_name)
    # > dev.sources
    ```

    References
    ----------

    * [Databricks Volume](https://docs.databricks.com/en/sql/language-manual/sql-ref-volumes.html)
    * [Pulumi Databricks Volume](https://www.pulumi.com/registry/packages/databricks/api-docs/volume/)
    """

    name: str
    catalog_name: str = None
    grant: Union[VolumeGrant, list[VolumeGrant]] = None
    grants: list[VolumeGrant] = None
    schema_name: str = None
    storage_location: str = None
    volume_type: Literal["MANAGED", "EXTERNAL"] = "MANAGED"

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        """Schema full name `{catalog_name}.{schema_name}`"""
        _id = ""
        if self.catalog_name:
            _id += self.catalog_name

        if self.schema_name:
            if _id == "":
                _id = self.schema_name
            else:
                _id += f".{self.schema_name}"

        return _id

    @property
    def full_name(self) -> str:
        """Volume full name `{catalog_name}.{schema_name}.{volume_name}`"""
        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """Table full name (catalog.schema.volume)"""
        return self.full_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - volume grants
        """
        resources = []

        # Volume grants
        resources += self.get_grants_additional_resources()

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Volume"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["grant", "grants"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_volume"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import typer

from laktory._version import VERSION

APP_NAME = "laktory-cli"

app = typer.Typer(
    pretty_exceptions_show_locals=False,
    help=f"Laktory {VERSION} CLI to preview and deploy resources",
)  # prevent display secret data


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    version: bool = typer.Option(
        False, "--version", "-v", help="Show laktory CLI version"
    ),
):
    """ """
    if version:
        print(f"Laktory CLI version {VERSION}")
        raise typer.Exit()



from typing import Literal

from laktory.models.basemodel import BaseModel


class MetastoreGrant(BaseModel):
    """
    Privileges granted to a principal and operating on a metastore

    Attributes
    ----------
    principal
        User, group or service principal name
    privileges
        List of allowed privileges

    References
    ----------
    * [privilege types](https://docs.databricks.com/en/sql/language-manual/sql-ref-privileges.html#privilege-types)
    """

    principal: str
    privileges: list[
        Literal[
            "CREATE_CATALOG",
            "CREATE_CONNECTION",
            "CREATE_EXTERNAL_LOCATION",
            "CREATE_PROVIDER",
            "CREATE_RECIPIENT",
            "CREATE_SHARE",
            "CREATE_STORAGE_CREDENTIAL",
            "LOCATION",
            "MANAGE_ALLOWLIST",
            "SET_SHARE_PERMISSION",
            "USE_MARKETPLACE_ASSETS",
            "USE_PROVIDER",
            "USE_RECIPIENT",
            "USE_SHARE",
        ]
    ]



import os

from laktory._settings import settings
from laktory.constants import CACHE_ROOT
from laktory.models.pipeline.pipelinechild import PipelineChild
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.workspacefile import WorkspaceFile


class PipelineRequirementsWorkspaceFile(WorkspaceFile, PipelineChild):
    """
    Workspace File storing pipeline python requirements. Default values for
    path and access controls. Forced value for source.

    Attributes
    ----------
    access_controls:
        List of file access controls
    path:
         Workspace filepath for the file. Overwrite `rootpath` and `dirpath`.
         Default value `{settings.workspace_laktory_root}pipelines/{pl_name}/requirements.txt`
    """

    source: str = "{pl_name}"
    access_controls: list[AccessControl] = [
        AccessControl(permission_level="CAN_READ", group_name="users")
    ]

    def update_from_parent(self):
        pl = self.parent_pipeline
        pl_name = pl.name
        self.source = os.path.join(CACHE_ROOT, f"tmp-{pl_name}-requirements.txt")
        if "{pl_name}" in self.path:
            self.path = (
                f"{settings.workspace_laktory_root}pipelines/{pl_name}/requirements.txt"
            )
        self.set_paths()

    def write_source(self):
        pl = self.parent_pipeline

        source = self.inject_vars_into_dump({"source": self.source})["source"]
        deps = self.inject_vars_into_dump({"deps": pl._dependencies})["deps"]

        with open(source, "w", newline="\n") as fp:
            fp.write("\n".join(deps))

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_type_id(self):
        return "workspace-file"



polars_installed = True
try:
    import polars
except ModuleNotFoundError:
    polars_installed = False


if polars_installed:
    from polars import DataFrame as PolarsDataFrame
    from polars import Expr as PolarsExpr
    from polars import LazyFrame as PolarsLazyFrame

    import laktory.polars.dataframe
    import laktory.polars.expressions
    from laktory.polars.datatypes import DATATYPES_MAP

    def is_polars_dataframe(df):
        """Check if dataframe is Polars DataFrame or Polars Lazy DataFrame"""
        if isinstance(df, PolarsDataFrame):
            return True

        if isinstance(df, PolarsLazyFrame):
            return True

        return False

else:
    # Mocks when polars is not installed

    class PolarsDataFrame:
        pass

    class PolarsLazyFrame:
        pass

    class PolarsExpr:
        pass

    def is_polars_dataframe(df):
        return False



from ._version import VERSION

__version__ = VERSION

# Import first
from ._settings import settings
from ._useragent import set_databricks_sdk_upstream

set_databricks_sdk_upstream()

import laktory._parsers
import laktory.models
import laktory.spark
import laktory.typing
import laktory.yaml

from ._logger import get_logger
from ._settings import Settings
from .version import show_version_info



import os
import subprocess
from typing import Union

from prompt_toolkit.validation import ValidationError
from prompt_toolkit.validation import Validator
from pydantic import BaseModel

from laktory._logger import get_logger
from laktory.constants import QUICKSTART_TEMPLATES
from laktory.constants import SUPPORTED_BACKENDS
from laktory.models.stacks.stack import Stack

logger = get_logger(__name__)
DIRPATH = os.path.dirname(__file__)


class TemplateValidator(Validator):
    def validate(self, document):
        text = document.text
        if text.lower() not in QUICKSTART_TEMPLATES:
            raise ValidationError(
                message=f"Please enter one of the supported template {QUICKSTART_TEMPLATES}",
                cursor_position=len(text),
            )  # Move cursor to end


class BackendValidator(Validator):
    def validate(self, document):
        text = document.text
        if text.lower() not in SUPPORTED_BACKENDS:
            raise ValidationError(
                message=f"Please enter one of the supported IaC backends {SUPPORTED_BACKENDS}",
                cursor_position=len(text),
            )  # Move cursor to end


class CLIController(BaseModel):
    stack_filepath: Union[str, None] = None
    env: Union[str, None] = None
    auto_approve: Union[bool, None] = False
    options_str: Union[str, None] = None
    stack: Union[Stack, None] = None

    def model_post_init(self, __context):
        super().model_post_init(__context)

        # Read stack
        if self.stack_filepath is None:
            self.stack_filepath = "./stack.yaml"
        logger.info(f"Reading stack from '{self.stack_filepath}'")
        with open(self.stack_filepath, "r", encoding="utf-8") as fp:
            self.stack = Stack.model_validate_yaml(fp)

        # Check environment
        if self.env is None:
            env_names = list(self.stack.environments.keys())
            if env_names:
                logger.warn(
                    f"Environment not specified, defaulting to first available ({env_names[0]})"
                )
                self.env = env_names[0]

    @property
    def backend(self) -> str:
        return self.stack.backend

    @property
    def organization(self) -> str:
        return self.stack.organization

    @property
    def pulumi_options(self):
        options = []
        if self.auto_approve:
            options += ["--yes"]

        if self.options_str:
            options += self.options_str.split(",")

        return options

    @property
    def terraform_options(self):
        options = []
        if self.auto_approve:
            options += ["-auto-approve"]

        if self.options_str:
            options += self.options_str.split(",")

        return options

    @property
    def pulumi_stack_name(self):
        return self.organization + "/" + self.env

    def pulumi_call(self, cmd):
        if self.pulumi_stack_name is None:
            raise ValueError("Argument `stack` must be specified with pulumi backend")

        pstack = self.stack.to_pulumi(env_name=self.env)
        getattr(pstack, cmd)(stack=self.pulumi_stack_name, flags=self.pulumi_options)

    def terraform_call(self, cmd):
        pstack = self.stack.to_terraform(env_name=self.env)
        getattr(pstack, cmd)(flags=self.terraform_options)


class Worker:
    def run(self, cmd, cwd=None, raise_exceptions=True):
        try:
            subprocess.run(
                cmd,
                cwd=cwd,
                check=True,
            )

        except Exception as e:
            _cmd = " ".join(cmd)
            if raise_exceptions:
                raise e
            else:
                print(f"An error occurred while executing '{_cmd}': {str(e)}")

                # Windows
                c1 = (
                    _cmd.startswith("terraform")
                    and "The system cannot find the file specified".lower()
                    in str(e).lower()
                )
                c2 = (
                    _cmd.startswith("pulumi")
                    and "The system cannot find the file specified".lower()
                    in str(e).lower()
                )

                # Mac/Linux
                c3 = "No such file or directory: 'terraform'".lower() in str(e).lower()
                c4 = "No such file or directory: 'pulumi'".lower() in str(e).lower()

                if c1 or c3:
                    print(
                        "Terraform is selected as IaC backend. Make sure it is installed and part of the PATH"
                    )
                elif c2 or c4:
                    print(
                        "Pulumi is selected as IaC backend. Make sure it is installed and part of the PATH"
                    )



import polars as pl
import sqlparse
from sqlparse.sql import Comparison
from sqlparse.tokens import Keyword


def _parse_token(token):
    # Numerical
    try:
        _ = float(token)
        if "." in token:
            return _
        else:
            return int(token)
    except (ValueError, TypeError):
        pass

    # String constant
    if token.startswith("'") or token.startswith('"'):
        return token[1:-1]

    # Structure
    if "." in token:
        names = token.split(".")
        col = pl.col(names[0])
        for n in names[1:]:
            col = col.struct.field(n)
        return col

    # Column
    return pl.col(token)


def _parse_compare(condition: str):
    # Remove whitespaces
    condition = condition.replace(" ", "")

    # Breakdown
    parsed = sqlparse.parse(condition)[0]
    comparison = [token for token in parsed.tokens if isinstance(token, Comparison)][0]
    left = comparison.left.value
    operator = comparison.tokens[1].value
    right = comparison.right.value

    # Build expression
    if operator == "<":
        return _parse_token(left) < _parse_token(right)
    elif operator == ">":
        return _parse_token(left) > _parse_token(right)
    elif operator == "<=":
        return _parse_token(left) <= _parse_token(right)
    elif operator == ">=":
        return _parse_token(left) >= _parse_token(right)
    elif operator == "==":
        return _parse_token(left) == _parse_token(right)
    elif operator == "!=":
        return _parse_token(left) != _parse_token(right)

    return condition


def sql_expr(sql: str) -> pl.Expr:
    """
    Parse SQL expression to polars expression(s) with support for nested
    structure

    Parameters
    ----------
    sql :
        SQL expression

    Returns
    -------
    :
        Polar expression

    ```py
    import polars as pl

    import laktory  # noqa: F401

    exp = pl.Expr.laktory.sql_expr("data.close")
    print(exp)
    # > col("data").struct.field_by_name(close)()
    exp = pl.Expr.laktory.sql_expr("data.close > 5.0")
    print(exp)
    # > [(col("data").struct.field_by_name(close)()) > (dyn float: 5.0)]
    ```
    """
    try:
        return pl.sql_expr(sql)

    except pl.exceptions.ComputeError:
        parsed = sqlparse.parse(sql)[0]
        expressions = []

        for token in parsed.tokens:
            if isinstance(token, Comparison):
                expressions += [_parse_compare(str(token))]
            elif token.ttype is Keyword and token.value.upper() in ["AND", "OR"]:
                expressions.append(token.value.upper())
            elif str(token).replace(" ", "") == "":
                pass
            else:
                expressions += [_parse_token(str(token))]

        expr = expressions[0]
        i = 1
        while i < len(expressions) - 1:
            if expressions[i] == "AND":
                expr = expr & expressions[i + 1]
            elif expressions[i] == "OR":
                expr = expr | expressions[i + 1]
            i += 2

        return expr



from typing import Any
from typing import Literal
from typing import Union

from pydantic import Field
from pydantic import model_validator

from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.cluster import Cluster
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class PipelineLibraryFile(BaseModel):
    """
    Pipeline Library File specifications

    Attributes
    ----------
    path:
        Workspace filepath
    """

    path: str


class PipelineLibraryNotebook(BaseModel):
    """
    Pipeline Library Notebook specifications

    Attributes
    ----------
    path:
        Workspace notebook filepath
    """

    path: str


class PipelineLibrary(BaseModel):
    """
    Pipeline Library specifications

    Attributes
    ----------
    file:
        File specifications
    notebook:
        Notebook specifications
    """

    file: str = None
    notebook: PipelineLibraryNotebook = None


class PipelineNotifications(BaseModel):
    """
    Pipeline Notifications specifications

    Attributes
    ----------
    alerts:
        Alert types
    recipients:
        List of user/group/service principal names
    """

    alerts: list[
        Literal[
            "on-update-success",
            "on-update-failure",
            "on-update-fatal-failure",
            "on-flow-failure",
        ]
    ]
    recipients: list[str]


class PipelineCluster(Cluster):
    """
    Pipeline Cluster. Same attributes as `laktory.models.Cluster`, except for

    * `autotermination_minutes`
    * `cluster_id`
    * `data_security_mode`
    * `enable_elastic_disk`
    * `idempotency_token`
    * `is_pinned`
    * `libraries`
    * `no_wait`
    * `node_type_id`
    * `runtime_engine`
    * `single_user_name`
    * `spark_version`

    that are not allowed.
    """

    autotermination_minutes: int = Field(None, exclude=True)
    cluster_id: str = Field(None, exclude=True)
    data_security_mode: str = Field(None, exclude=True)
    enable_elastic_disk: bool = Field(None, exclude=True)
    idempotency_token: str = Field(None, exclude=True)
    is_pinned: bool = Field(None, exclude=True)
    libraries: list[Any] = Field(None, exclude=True)
    node_type_id: str = None
    no_wait: bool = Field(None, exclude=True)
    runtime_engine: str = Field(None, exclude=True)
    single_user_name: str = Field(None, exclude=True)
    spark_version: str = Field(None, exclude=True)

    @model_validator(mode="after")
    def excluded_fields(self) -> Any:
        for f in [
            "autotermination_minutes",
            "cluster_id",
            "data_security_mode",
            "enable_elastic_disk",
            "idempotency_token",
            "is_pinned",
            "libraries",
            "no_wait",
            "runtime_engine",
            "single_user_name",
            "spark_version",
        ]:
            if getattr(self, f, None) not in [None, [], {}]:
                raise ValueError(f"Field {f} should be null")

        return self


class DLTPipeline(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Delta Live Tables (DLT) Pipeline

    Attributes
    ----------
    access_controls:
        Pipeline access controls
    allow_duplicate_names:
        If `False`, deployment will fail if name conflicts with that of another pipeline.
    catalog:
        Name of the unity catalog storing the pipeline tables
    channel:
        Name of the release channel for Spark version used by DLT pipeline.
    clusters:
        Clusters to run the pipeline. If none is specified, pipelines will automatically select a default cluster
        configuration for the pipeline.
    configuration:
         List of values to apply to the entire pipeline. Elements must be formatted as key:value pairs
    continuous:
        If `True`, the pipeline is run continuously.
    development:
        If `True` the pipeline is run in development mode
    edition:
        Name of the product edition
    libraries:
        Specifies pipeline code (notebooks) and required artifacts.
    name:
        Pipeline name
    name_prefix:
        Prefix added to the DLT pipeline name
    name_suffix:
        Suffix added to the DLT pipeline name
    notifications:
        Notifications specifications
    photon:
        If `True`, Photon engine enabled.
    serverless:
        If `True`, serverless is enabled
    storage:
        A location on DBFS or cloud storage where output data and metadata required for pipeline execution are stored.
        By default, tables are stored in a subdirectory of this location. Change of this parameter forces recreation
        of the pipeline. (Conflicts with `catalog`).
    target:
        The name of a database (in either the Hive metastore or in a UC catalog) for persisting pipeline output data.
        Configuring the target setting allows you to view and query the pipeline output data from the Databricks UI.

    Examples
    --------
    Assuming the configuration yaml file
    ```py
    import io

    from laktory import models

    # Define pipeline
    pipeline_yaml = '''
    name: pl-stock-prices

    catalog: dev
    target: finance

    clusters:
      - name : default
        node_type_id: Standard_DS3_v2
        autoscale:
          min_workers: 1
          max_workers: 2

    libraries:
      - notebook:
          path: /pipelines/dlt_brz_template.py
      - notebook:
          path: /pipelines/dlt_slv_template.py
      - notebook:
          path: /pipelines/dlt_gld_stock_performances.py

    access_controls:
      - group_name: account users
        permission_level: CAN_VIEW
      - group_name: role-engineers
        permission_level: CAN_RUN

    '''
    pipeline = models.resources.databricks.DLTPipeline.model_validate_yaml(
        io.StringIO(pipeline_yaml)
    )
    ```

    References
    ----------

    * [Databricks Pipeline](https://docs.databricks.com/api/workspace/pipelines/create)
    * [Pulumi Databricks Pipeline](https://www.pulumi.com/registry/packages/databricks/api-docs/pipeline/)
    """

    access_controls: list[AccessControl] = []
    allow_duplicate_names: bool = None
    catalog: Union[str, None] = None
    channel: Literal["CURRENT", "PREVIEW"] = "PREVIEW"
    clusters: list[PipelineCluster] = []
    configuration: dict[str, str] = {}
    continuous: bool = None
    development: bool = None
    edition: Literal["CORE", "PRO", "ADVANCED"] = None
    # filters
    libraries: list[PipelineLibrary] = None
    name: str
    name_prefix: str = None
    name_suffix: str = None
    notifications: list[PipelineNotifications] = []
    photon: bool = None
    serverless: bool = None
    storage: str = None
    target: str = None

    @model_validator(mode="after")
    def update_name(self) -> Any:
        with self.validate_assignment_disabled():
            if self.name_prefix:
                self.name = self.name_prefix + self.name
                self.name_prefix = ""
            if self.name_suffix:
                self.name = self.name + self.name_suffix
                self.name_suffix = ""
        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_type_id(self) -> str:
        """
        dlt
        """
        return "dlt-pipeline"

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []

        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    pipeline_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Pipeline"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return {
            "access_controls": True,
            "clusters": {"__all__": {"access_controls"}},
            "name_prefix": True,
            "name_suffix": True,
        }

    @property
    def pulumi_properties(self):
        d = super().pulumi_properties
        k = "clusters"
        if k in d:
            _clusters = []
            for c in d[k]:
                c["label"] = c.pop("name")
                _clusters += [c]
            d[k] = _clusters
        return d

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_pipeline"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes

    @property
    def terraform_properties(self) -> dict:
        d = super().terraform_properties
        k = "cluster"
        if k in d:
            _clusters = []
            for c in d[k]:
                c["label"] = c.pop("name")
                _clusters += [c]
            d[k] = _clusters
        return d



import json
import os.path
from pathlib import Path
from typing import Any
from typing import Literal
from typing import Union

from pydantic import ConfigDict
from pydantic import Field
from pydantic import field_validator
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.models.datasources.basedatasource import BaseDataSource
from laktory.polars import PolarsLazyFrame
from laktory.spark import SparkDataFrame

logger = get_logger(__name__)


class FileDataSource(BaseDataSource):
    """
    Data source using disk files, such as data events (json/csv) and
    dataframe parquets. It is generally used in the context of a data pipeline.

    Attributes
    ----------
    format:
        Format of the data files
    read_options:
        Other options passed to `spark.read.options`
    schema:
        Target schema specified as a list of columns, as a dict or a json
        serialization. Only used when reading data from non-strongly typed
        files such as JSON or csv files.
    schema_location:
        Path for schema inference when reading data as a stream. If `None`,
        parent directory of `path` is used.

    Examples
    ---------
    ```python
    from laktory import models

    source = models.FileDataSource(
        path="/Volumes/sources/landing/events/yahoo-finance/stock_price",
        format="JSON",
        as_stream=False,
    )
    # df = source.read(spark)

    # With Explicit Schema
    source = models.FileDataSource(
        path="/Volumes/sources/landing/events/yahoo-finance/stock_price",
        format="JSON",
        as_stream=False,
        schema=[
            {"name": "description", "type": "string", "nullable": True},
            {"name": "close", "type": "double", "nullable": False},
        ],
    )
    # df = source.read(spark)
    ```
    """

    model_config = ConfigDict(populate_by_name=True)

    format: Literal[
        "AVRO",
        "BINARYFILE",
        "CSV",
        "DELTA",
        "EXCEL",
        "JSON",
        "JSONL",
        "NDJSON",
        "ORC",
        "PARQUET",
        "TEXT",
        "XML",
    ] = "JSONL"
    path: str
    read_options: dict[str, Any] = {}
    schema_definition: Union[str, dict, list] = Field(None, validation_alias="schema")
    schema_location: str = None

    @field_validator("path", "schema_location", mode="before")
    @classmethod
    def posixpath_to_string(cls, value: Any) -> Any:
        if isinstance(value, Path):
            value = str(value)
        return value

    @model_validator(mode="after")
    def options(self) -> Any:
        if self.dataframe_backend == "SPARK":
            if self.format in [
                "EXCEL",
            ]:
                raise ValueError(f"'{self.format}' format is not supported with Spark")

        elif self.df_backend == "POLARS":
            if self.format in [
                "AVRO",
                "BINARYFILE",
                "ORC",
                "TEXT",
                "XML",
            ]:
                raise ValueError(f"'{self.format}' format is not supported with Polars")

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self):
        return str(self.path)

    @property
    def _schema(self):
        schema = self.schema_definition
        if schema is None:
            return schema
        if isinstance(schema, list):
            schema = {"fields": schema, "type": "struct"}

        if isinstance(schema, str) and '"fields":[' in schema.replace("'", '"').replace(
            " ", ""
        ):
            schema = json.loads(schema)

        # DDL format
        if isinstance(schema, str):
            pass

        # Spark Struct format
        elif isinstance(schema, dict):
            import pyspark.sql.types as T

            schema = T.StructType.fromJson(schema)

        return schema

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _read_spark(self, spark) -> SparkDataFrame:
        _options = {}
        _mode = "stream"

        # JSON
        _format = self.format
        if self.format in ["NDJSON", "JSONL"]:
            _format = "JSON"
            _options["multiline"] = False
        elif self.format == "JSON":
            _format = "JSON"
            _options["multiline"] = True

        # CSV
        if self.format == "CSV":
            _options["header"] = True

        if self.as_stream:
            _mode = "stream"

            if _format == "DELTA":
                reader = spark.readStream.format(_format.lower())

            else:
                reader = spark.readStream.format("cloudFiles")
                _options["cloudFiles.format"] = _format

                if self._schema:
                    reader = reader.schema(self._schema)
                else:
                    schema_location = self.schema_location
                    if schema_location is None:
                        schema_location = os.path.dirname(self.path)
                    _options["cloudFiles.schemaLocation"] = schema_location

            if self._schema is None:
                _options["cloudFiles.inferColumnTypes"] = True
                _options["cloudFiles.schemaEvolutionMode"] = "addNewColumns"

        else:
            _mode = "static"
            reader = spark.read.format(_format.lower())
            if self._schema:
                reader = reader.schema(self._schema)

        # User Options
        _options["mergeSchema"] = True
        _options["recursiveFileLookup"] = True
        if self.read_options:
            for k, v in self.read_options.items():
                _options[k] = v

        reader = reader.options(**_options)

        # Load
        logger.info(f"Reading {self._id} as {_mode} and options {_options}")
        if self._schema:
            schema_str = self._schema
            if hasattr(schema_str, "simpleString"):
                schema_str = schema_str.simpleString()
            else:
                schema_str = str(schema_str)
            logger.info(f"Expected schema: {schema_str}")
        df = reader.load(self.path)

        return df

    def _read_polars(self) -> PolarsLazyFrame:
        import polars as pl

        if self.as_stream:
            raise ValueError(
                "Streaming read not supported with Pandas DataFrame. Please switch to Spark"
            )

        logger.info(f"Reading {self._id} as static")

        if self.format.lower() == "csv":
            df = pl.scan_csv(self.path, **self.read_options)

        elif self.format.lower() == "delta":
            df = pl.scan_delta(self.path, **self.read_options)

        elif self.format.lower() == "excel":
            df = pl.read_excel(self.path, **self.read_options)

        elif self.format.lower() == "json":
            df = pl.read_json(self.path, **self.read_options)

        elif self.format.lower() in ["jsonl", "ndjson"]:
            df = pl.scan_ndjson(self.path, **self.read_options)

        elif self.format.lower() == "parquet":
            df = pl.scan_parquet(self.path, **self.read_options)

        else:
            raise ValueError(f"Format '{self.format}' is not supported.")

        if isinstance(df, pl.DataFrame):
            df = df.lazy()

        return df



from __future__ import annotations

import abc
import re
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Literal
from typing import Union

from pydantic import field_validator
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.constants import SUPPORTED_DATATYPES
from laktory.models.basemodel import BaseModel
from laktory.models.dataframecolumnexpression import DataFrameColumnExpression
from laktory.models.pipeline.pipelinechild import PipelineChild
from laktory.polars import PolarsDataFrame
from laktory.polars import PolarsExpr
from laktory.typing import AnyDataFrame

if TYPE_CHECKING:
    from laktory.models.datasources.basedatasource import BaseDataSource
    from laktory.models.datasources.pipelinenodedatasource import PipelineNodeDataSource


logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Helper Classes                                                              #
# --------------------------------------------------------------------------- #


class BaseChainNodeFuncArg(BaseModel, PipelineChild):
    value: Union[Any]

    @field_validator("value")
    def value_to_data_source(cls, v: Any) -> Any:
        """
        Data source can't be set as an expected value type as it would create
        a circular dependency. Instead, we check at validation if the value
        could be instantiated as a DataSource object.
        """
        from laktory.models.datasources import classes

        for c in classes:
            try:
                v = c(**v)
                break
            except:  # noqa: E722
                pass

        return v

    @abc.abstractmethod
    def eval(self):
        raise NotImplementedError()

    def signature(self):
        from laktory.models.datasources import FileDataSource
        from laktory.models.datasources import PipelineNodeDataSource
        from laktory.models.datasources import TableDataSource

        if isinstance(self.value, PipelineNodeDataSource):
            return f"node.{self.value.node_name}"
        elif isinstance(self.value, FileDataSource):
            return f"file {self.value.path}"
        elif isinstance(self.value, TableDataSource):
            return f"table {self.value.full_name}"
        return str(self.value)


class ChainNodeColumn(BaseModel, PipelineChild):
    expr: Union[str, DataFrameColumnExpression]
    name: str
    type: Union[str, None] = "string"
    unit: Union[str, None] = None

    @model_validator(mode="after")
    def parse_expr(self) -> Any:
        if isinstance(self.expr, str):
            self.expr = DataFrameColumnExpression(value=self.expr)
        return self

    @field_validator("type")
    def check_type(cls, v: str) -> str:
        if v is None or "<" in v:
            return v
        else:
            if v not in SUPPORTED_DATATYPES:
                raise ValueError(
                    f"Type {v} is not supported. Select one of {SUPPORTED_DATATYPES}"
                )
        return v

    def eval(self, udfs=None, dataframe_backend=None):
        return self.expr.eval(udfs=udfs, dataframe_backend=dataframe_backend)


class BaseChainNodeSQLExpr(BaseModel, PipelineChild):
    """
    Chain node SQL expression

    Attributes
    ----------
    expr:
        SQL expression
    """

    expr: str
    _data_sources: list[PipelineNodeDataSource] = None

    def parsed_expr(self, df_id="df", view=False) -> list[str]:
        from laktory.models.datasources.pipelinenodedatasource import (
            PipelineNodeDataSource,
        )
        from laktory.models.datasources.tabledatasource import TableDataSource

        expr = self.expr
        if view:
            pl_node = self.parent_pipeline_node

            if pl_node and pl_node.source:
                source = pl_node.source
                if isinstance(source, TableDataSource):
                    full_name = source.full_name
                elif isinstance(source, PipelineNodeDataSource):
                    full_name = source.sink_table_full_name
                else:
                    raise ValueError(
                        "VIEW sink only supports Table or Pipeline Node with Table sink data sources"
                    )
                expr = expr.replace("{df}", full_name)

            pl = self.parent_pipeline
            if pl:
                from laktory.models.datasinks.tabledatasink import TableDataSink

                pattern = r"\{nodes\.(.*?)\}"
                matches = re.findall(pattern, expr)
                for m in matches:
                    if m not in pl.nodes_dict:
                        raise ValueError(
                            f"Node '{m}' is not available from pipeline '{pl.name}'"
                        )
                    sink = pl.nodes_dict[m].primary_sink
                    if not isinstance(sink, TableDataSink):
                        raise ValueError(
                            f"Node '{m}' used in view creation does not have a Table sink"
                        )
                    expr = expr.replace("{nodes." + m + "}", sink.full_name)

            return expr

        expr = expr.replace("{df}", df_id)
        pattern = r"\{nodes\.(.*?)\}"
        matches = re.findall(pattern, expr)
        for m in matches:
            expr = expr.replace("{nodes." + m + "}", f"nodes__{m}")

        return expr.split(";")

    @property
    def upstream_node_names(self) -> list[str]:
        if self.expr is None:
            return []

        names = []

        pattern = r"\{nodes\.(.*?)\}"
        matches = re.findall(pattern, self.expr)
        for m in matches:
            names += [m]

        return names

    @property
    def data_sources(self) -> list[PipelineNodeDataSource]:
        if self._data_sources is None:
            if self.expr is None:
                return []

            from laktory.models.datasources.pipelinenodedatasource import (
                PipelineNodeDataSource,
            )

            sources = []

            pattern = r"\{nodes\.(.*?)\}"
            matches = re.findall(pattern, self.expr)
            for m in matches:
                sources += [PipelineNodeDataSource(node_name=m)]

            self._data_sources = sources

        return self._data_sources

    def eval(self, df):
        raise NotImplementedError()


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class BaseChainNode(BaseModel, PipelineChild):
    dataframe_backend: Literal["SPARK", "POLARS", None] = None
    func_args: list[Union[Any]] = []
    func_kwargs: dict[str, Union[Any]] = {}
    func_name: Union[str, None] = None
    sql_expr: Union[str, None] = None
    with_column: Union[ChainNodeColumn, None] = None
    with_columns: Union[list[ChainNodeColumn], None] = []
    _parsed_func_args: list = None
    _parsed_func_kwargs: dict = None
    _parsed_sql_expr: BaseChainNodeSQLExpr = None

    @model_validator(mode="after")
    def selected_flow(self) -> Any:
        if len(self._with_columns) > 0:
            if self.func_name:
                raise ValueError(
                    "`func_name` should not be set when using `with_column`"
                )
            if self.sql_expr:
                raise ValueError(
                    "`sql_expr` should not be set when using `with_column`"
                )
        else:
            if self.func_name and self.sql_expr:
                raise ValueError(
                    "Only one of `func_name` and `sql_expr` should be set."
                )
            if not (self.func_name or self.sql_expr):
                raise ValueError(
                    "Either `func_name`, `sql_expr` or `with_column` should be set when using"
                )

        return self

    # ----------------------------------------------------------------------- #
    # Id                                                                      #
    # ----------------------------------------------------------------------- #

    @property
    def id(self):
        if self.is_column:
            return "-".join([c.name for c in self.with_columns])
        return "df"

    # ----------------------------------------------------------------------- #
    # Children                                                                #
    # ----------------------------------------------------------------------- #

    @property
    def child_attribute_names(self):
        return [
            "data_sources",
            "_parsed_func_args",
            "_parsed_func_kwargs",
            "_parsed_sql_expr",
        ]

    # ----------------------------------------------------------------------- #
    # Columns Creation                                                        #
    # ----------------------------------------------------------------------- #

    @property
    def _with_columns(self) -> list[ChainNodeColumn]:
        with_columns = [c for c in self.with_columns]
        if self.with_column:
            with_columns += [self.with_column]
        return with_columns

    @property
    def is_column(self):
        return len(self._with_columns) > 0

    # ----------------------------------------------------------------------- #
    # Data Sources                                                            #
    # ----------------------------------------------------------------------- #

    @property
    def data_sources(self) -> list[BaseDataSource]:
        """Get all sources feeding the Chain Node"""

        from laktory.models.datasources.basedatasource import BaseDataSource

        sources = []
        for a in self.parsed_func_args:
            if isinstance(a.value, BaseDataSource):
                sources += [a.value]
        for a in self.parsed_func_kwargs.values():
            if isinstance(a.value, BaseDataSource):
                sources += [a.value]

        if self.sql_expr:
            sources += self.parsed_sql_expr.data_sources

        return sources

    # ----------------------------------------------------------------------- #
    # Upstream Nodes                                                          #
    # ----------------------------------------------------------------------- #

    @property
    def upstream_node_names(self) -> list[str]:
        """Pipeline node names required to apply transformer node."""

        from laktory.models.datasources.pipelinenodedatasource import (
            PipelineNodeDataSource,
        )

        names = []
        for a in self.parsed_func_args:
            if isinstance(a.value, PipelineNodeDataSource):
                names += [a.value.node_name]
        for a in self.parsed_func_kwargs.values():
            if isinstance(a.value, PipelineNodeDataSource):
                names += [a.value.node_name]

        if self.sql_expr:
            names += self.parsed_sql_expr.upstream_node_names

        return names

    # ----------------------------------------------------------------------- #
    # Execution                                                               #
    # ----------------------------------------------------------------------- #

    @abc.abstractmethod
    def execute(
        self,
        df: AnyDataFrame,
        udfs: list[Callable[[...], Union[PolarsExpr, PolarsDataFrame]]] = None,
        return_col: bool = False,
    ) -> Union[AnyDataFrame]:
        raise NotImplementedError()

    def get_view_definition(self):
        return self._parsed_sql_expr.parsed_expr(view=True)



import os
import shutil
from pathlib import Path
from typing import Any
from typing import Literal
from typing import Union

from pydantic import field_validator
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.models.datasinks.basedatasink import BaseDataSink
from laktory.models.datasources.filedatasource import FileDataSource
from laktory.polars import PolarsDataFrame
from laktory.polars import PolarsLazyFrame
from laktory.spark import SparkDataFrame

logger = get_logger(__name__)


class FileDataSink(BaseDataSink):
    """
    Disk file(s) data sink such as csv, parquet or Delta Table.

    Attributes
    ----------
    checkpoint_location:
        Path to which the checkpoint file for streaming dataframe should
        be written. If `None`, parent directory of `path` is used.
    format:
        Format of the data files
    path:
        Path to which the DataFrame needs to be written.

    Examples
    ---------
    ```python
    import pandas as pd

    from laktory import models

    df = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["AAPL", "GOOGL"],
                "price": [200.0, 205.0],
                "tstamp": ["2023-09-01", "2023-09-01"],
            }
        )
    )

    sink = models.FileDataSink(
        path="/Volumes/sources/landing/events/yahoo-finance/stock_price",
        format="PARQUET",
        mode="OVERWRITE",
    )
    # sink.write(df)

    # Sink with Change Data Capture processing
    sink = models.FileDataSink(
        path="/Volumes/sources/landing/events/yahoo-finance/stock_price",
        format="DELTA",
        mode="MERGE",
        merge_cdc_options={
            "scd_type": 1,
            "primary_keys": ["symbol", "tstamp"],
        },
    )
    # sink.write(df)
    ```
    """

    checkpoint_location: Union[str, None] = None
    format: Literal["CSV", "PARQUET", "DELTA", "JSON", "EXCEL"] = "DELTA"
    path: str

    @field_validator("path", mode="before")
    @classmethod
    def posixpath_to_string(cls, value: Any) -> Any:
        if isinstance(value, Path):
            value = str(value)
        return value

    @model_validator(mode="after")
    def merge_and_format(self) -> Any:
        if self.mode == "MERGE":
            if self.format.lower() not in ["delta"]:
                raise ValueError(
                    "Merge write mode is only supported for 'DELTA' `format`"
                )

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self):
        return str(self.path)

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def _write_spark(self, df: SparkDataFrame, mode=None, full_refresh=False) -> None:
        if self.format in ["EXCEL"]:
            raise ValueError(f"'{self.format}' format is not supported with Spark")

        # Set mode
        if mode is None:
            mode = self.mode

        if mode.lower() == "merge":
            self.merge_cdc_options.execute(source=df)
            return

        # Full Refresh
        if full_refresh or not self.exists(spark=df.sparkSession):
            if df.isStreaming:
                pass
                # .is_aggregate() method seems unreliable. Disabling for now.
                # if df.laktory.is_aggregate():
                #     logger.info(
                #         "Full refresh or initial load. Switching to COMPLETE mode."
                #     )
                #     mode = "COMPLETE"
            else:
                logger.info(
                    "Full refresh or initial load. Switching to OVERWRITE mode."
                )
                mode = "OVERWRITE"

        # Default Options
        _options = {"mergeSchema": "true", "overwriteSchema": "false"}
        if mode in ["OVERWRITE", "COMPLETE"]:
            _options["mergeSchema"] = "false"
            _options["overwriteSchema"] = "true"
        if df.isStreaming:
            _options["checkpointLocation"] = self._checkpoint_location

        # User Options
        for k, v in self.write_options.items():
            _options[k] = v

        if df.isStreaming:
            logger.info(
                f"Writing df as stream {self.format} to {self.path} with mode {mode} and options {_options}"
            )
            writer = (
                df.writeStream.format(self.format.lower())
                .outputMode(mode)
                .trigger(availableNow=True)  # TODO: Add option for trigger?
                .options(**_options)
            )
            if self.cluster_by:
                writer = writer.clusterBy(*self.cluster_by)
            query = writer.start(self.path)
            query.awaitTermination()

        else:
            logger.info(
                f"Writing df as static {self.format} to {self.path} with mode {mode} and options {_options}"
            )
            writer = df.write.mode(mode).format(self.format.lower()).options(**_options)
            if self.cluster_by:
                writer = writer.clusterBy(*self.cluster_by)
            writer.save(self.path)

    def _write_polars(self, df: PolarsDataFrame, mode=None, full_refresh=False) -> None:
        isStreaming = False

        if self.format != "DELTA":
            if mode:
                raise ValueError(
                    "'mode' configuration with Polars only supported by 'DELTA' format"
                )
        else:
            if full_refresh or not self.exists():
                mode = "OVERWRITE"

            if not mode:
                raise ValueError(
                    "'mode' configuration required with Polars 'DELTA' format"
                )

        if isStreaming:
            logger.info(
                f"Writing df as stream {self.format} to {self.path} with mode {mode}"
            )

        else:
            logger.info(
                f"Writing df as static {self.format} to {self.path} with mode {mode}"
            )

        if isinstance(df, PolarsLazyFrame):
            df = df.collect()

        if self.format.lower() == "csv":
            df.write_csv(self.path, **self.write_options)
        elif self.format.lower() == "delta":
            df.write_delta(self.path, mode=mode, **self.write_options)
        elif self.format.lower() == "excel":
            df.write_excel(self.path, **self.write_options)
        elif self.format.lower() == "json":
            df.write_json(self.path, **self.write_options)
        elif self.format.lower() == "parquet":
            df.write_parquet(self.path, **self.write_options)

    # ----------------------------------------------------------------------- #
    # Purge                                                                   #
    # ----------------------------------------------------------------------- #

    def purge(self, spark=None):
        """
        Delete sink data and checkpoints
        """
        # TODO: Now that sink switch to overwrite when sink does not exists or when
        # a full refresh is requested, the purge method should not delete the data
        # by default, but only the checkpoints. Also consider truncating the table
        # instead of dropping it.

        # Remove Data
        if os.path.exists(self.path):
            is_dir = os.path.isdir(self.path)
            if is_dir:
                logger.info(f"Deleting data dir {self.path}")
                shutil.rmtree(self.path)
            else:
                logger.info(f"Deleting data file {self.path}")
                os.remove(self.path)

        # TODO: Add support for Databricks dbfs / workspace / Volume?

        # Remove Checkpoint
        self._purge_checkpoint(spark=spark)

    # ----------------------------------------------------------------------- #
    # Source                                                                  #
    # ----------------------------------------------------------------------- #

    def as_source(self, as_stream: bool = None) -> FileDataSource:
        """
        Generate a file data source with the same path as the sink.

        Parameters
        ----------
        as_stream:
            If `True`, sink will be read as stream.

        Returns
        -------
        :
            File Data Source
        """

        source = FileDataSource(
            path=self.path,
            format=self.format,
        )

        if as_stream:
            source.as_stream = as_stream

        if self.dataframe_backend:
            source.dataframe_backend = self.dataframe_backend
        source.parent = self.parent

        return source



from typing import Literal
from typing import Union

from laktory._logger import get_logger
from laktory.models.transformers.basechain import BaseChain
from laktory.models.transformers.polarschainnode import PolarsChainNode

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class PolarsChain(BaseChain):
    """
    The `PolarsChain` class defines a series of Polars transformation to be
    applied to a dataframe. Each transformation is expressed as a node
    (`PolarsChainNode` object) that, upon execution, returns a new dataframe.
    Each node is executed sequentially in the provided order. A node may also
    be another `PolarsChain`.

    Attributes
    ----------
    dataframe_backend:
        Differentiator to select dataframe chain type
    nodes:
        The list of transformations to be executed.

    Examples
    --------
    ```py
    import polars as pl

    from laktory import models

    df0 = pl.DataFrame({"x": [1, 2, 3]})

    # Build Chain
    sc = models.PolarsChain(
        nodes=[
            {
                "with_column": {
                    "name": "cos_x",
                    "type": "double",
                    "expr": "pl.col('x').cos()",
                },
            },
            {
                "nodes": [
                    {
                        "func_name": "rename",
                        "func_args": [
                            {"x": "x_tmp"},
                        ],
                    },
                    {
                        "with_column": {
                            "name": "x2",
                            "type": "double",
                            "expr": "pl.col('x_tmp').sqrt()",
                        },
                    },
                ],
            },
            {
                "func_name": "drop",
                "func_args": [
                    "x_tmp",
                ],
            },
        ],
    )

    # Execute Chain
    df = sc.execute(df0)

    # Print result
    print(df.glimpse(return_as_string=True))
    '''
    Rows: 3
    Columns: 2
    $ cos_x <f64> 0.5403023058681398, -0.4161468365471424, -0.9899924966004454
    $ x2    <f64> 1.0, 1.4142135623730951, 1.7320508075688772
    '''
    ```
    """

    dataframe_backend: Literal["POLARS"] = "POLARS"
    nodes: list[Union[PolarsChainNode, "PolarsChain"]]
    _columns: list[list[str]] = []


PolarsChain.model_rebuild()



from typing import Union

import polars as pl
from pydantic import BaseModel

from laktory._logger import get_logger

logger = get_logger(__name__)


class OrderBy(BaseModel):
    sql_expression: str
    desc: bool = False


def window_filter(
    df,
    partition_by: Union[list[str], None],
    order_by: Union[list[OrderBy], None] = None,
    drop_row_index: bool = True,
    row_index_name: str = "_row_index",
    rows_to_keep: int = 1,
) -> pl.DataFrame:
    """
    Apply spark window-based filtering

    Parameters
    ----------
    df:
        DataFrame
    partition_by
        Defines the columns used for grouping into Windows
    order_by:
        Defines the column used for sorting before dropping rows
    drop_row_index:
        If `True`, the row index column is dropped
    row_index_name:
        Group-specific and sorted rows index name
    rows_to_keep:
        How many rows to keep per window


    Examples
    --------
    ```py
    import polars as pl

    import laktory  # noqa: F401

    df0 = pl.DataFrame(
        [
            ["2023-01-01T00:00:00Z", "APPL", 200.0],
            ["2023-01-02T00:00:00Z", "APPL", 202.0],
            ["2023-01-03T00:00:00Z", "APPL", 201.5],
            ["2023-01-01T00:00:00Z", "GOOL", 200.0],
            ["2023-01-02T00:00:00Z", "GOOL", 202.0],
            ["2023-01-03T00:00:00Z", "GOOL", 201.5],
        ],
        ["created_at", "symbol", "price"],
    ).with_columns(pl.col("created_at").cast(pl.Datetime))

    df = df0.laktory.window_filter(
        partition_by=["symbol"],
        order_by=[
            {"sql_expression": "created_at", "desc": True},
        ],
        drop_row_index=False,
        rows_to_keep=1,
    )

    print(df.glimpse(return_as_string=True))
    '''
    Rows: 2
    Columns: 4
    $ created_at <datetime[μs]> 2023-01-03 00:00:00, 2023-01-03 00:00:00
    $ symbol              <str> 'APPL', 'GOOL'
    $ price               <f64> 201.5, 201.5
    $ _row_index          <u32> 1, 1
    '''
    ```

    References
    ----------

    * [polars window](https://docs.pola.rs/user-guide/expressions/window/)
    """

    # Row Number
    e = pl.Expr.laktory.row_number()

    # Order by
    if order_by:
        bys = []
        descs = []
        for o in order_by:
            if not isinstance(o, OrderBy):
                o = OrderBy(**o)
            bys += [o.sql_expression]
            descs += [o.desc]
        e = e.sort_by(by=bys, descending=descs)

    # Partition By
    e = e.over(*partition_by)

    # Set rows index
    df = df.with_columns(**{row_index_name: e})

    # Filter
    df = df.filter(pl.col(row_index_name) <= rows_to_keep)
    if drop_row_index:
        df = df.drop(row_index_name)

    return df


if __name__ == "__main__":
    import polars as pl

    import laktory  # noqa: F401

    df0 = pl.DataFrame(
        [
            ["2023-01-01T00:00:00Z", "APPL", 200.0],
            ["2023-01-02T00:00:00Z", "APPL", 202.0],
            ["2023-01-03T00:00:00Z", "APPL", 201.5],
            ["2023-01-01T00:00:00Z", "GOOL", 200.0],
            ["2023-01-02T00:00:00Z", "GOOL", 202.0],
            ["2023-01-03T00:00:00Z", "GOOL", 201.5],
        ],
        ["created_at", "symbol", "price"],
    ).with_columns(pl.col("created_at").cast(pl.Datetime))

    df = df0.laktory.window_filter(
        partition_by=["symbol"],
        order_by=[
            {"sql_expression": "created_at", "desc": True},
        ],
        drop_row_index=False,
        rows_to_keep=1,
    )

    print(df.glimpse(return_as_string=True))



from functools import wraps

import polars as pl

from laktory.polars.expressions.datetime import current_timestamp
from laktory.polars.expressions.logical import compare
from laktory.polars.expressions.math import roundp

# from laktory.polars.expressions.sort import coalesce
from laktory.polars.expressions.sort import row_number
from laktory.polars.expressions.sql import sql_expr
from laktory.polars.expressions.string import string_split
from laktory.polars.expressions.string import uuid
from laktory.polars.expressions.units import convert_units


def _parse_args(args):
    """
    Polars expressions are designed to be chained to each other. When an
    extension expression is called, the calling expression is stored as
    `_expr` attribute. To support both chained operations as
    `pl.col("x").laktory.roundp(p=2.0)`
    and stand alone operations as
    `pl.Expr.laktory.roundp(pl.col("x"), p=2.0)`

    we check the value first argument received in an expression and return
    value._expr when appropriate.
    """
    if len(args) > 0 and isinstance(args[0], LaktoryExpression):
        args = list(args)
        args[0] = args[0]._expr
        return tuple(args)
    return args


@pl.api.register_expr_namespace("laktory")
class LaktoryExpression:
    def __init__(self, expr: pl.Expr):
        self._expr = expr

    # @wraps(coalesce)
    # def coalesce(*args, **kwargs):
    #     return coalesce(*_parse_args(args), **kwargs)

    @wraps(compare)
    def compare(*args, **kwargs):
        return compare(*_parse_args(args), **kwargs)

    @wraps(convert_units)
    def convert_units(*args, **kwargs):
        return convert_units(*_parse_args(args), **kwargs)

    @wraps(current_timestamp)
    def current_timestamp(*args, **kwargs):
        return current_timestamp(*_parse_args(args), **kwargs)

    @wraps(roundp)
    def roundp(*args, **kwargs):
        return roundp(*_parse_args(args), **kwargs)

    @wraps(row_number)
    def row_number(*args, **kwargs):
        return row_number(*_parse_args(args), **kwargs)

    @wraps(sql_expr)
    def sql_expr(*args, **kwargs):
        return sql_expr(*_parse_args(args), **kwargs)

    @wraps(string_split)
    def string_split(*args, **kwargs):
        return string_split(*_parse_args(args), **kwargs)

    @wraps(uuid)
    def uuid(*args, **kwargs):
        return uuid(*_parse_args(args), **kwargs)


pl.expr.laktory = pl.Expr.laktory



import re
from typing import Any
from typing import Literal

from pydantic import model_validator

from laktory._logger import get_logger
from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.typing import AnyDataFrameColumn

logger = get_logger(__name__)


class DataFrameColumnExpression(BaseModel):
    """
    DataFrame Column Expression supporting SQL statements or string
    representation of DataFrame API expression.

    Attributes
    ----------
    value:
        String representation
    type:
        Expression type: DF or SQL. If `None` is specified, type is guessed.

    Examples
    --------
    ```py
    from laktory import models

    e1 = models.DataFrameColumnExpression(
        value="MAX(close)",
    )
    print(e1.eval())
    # > Column<'MAX(close)'>

    e2 = models.DataFrameColumnExpression(
        value="F.abs('close')",
    )
    print(e2.eval())
    # > Column<'abs(close)'>
    ```
    """

    value: str
    type: Literal["SQL", "DF"] = None

    @model_validator(mode="after")
    def guess_type(self) -> Any:
        if self.type:
            return self

        expr_clean = self.value.strip().replace("\n", " ")

        type = "SQL"
        if re.findall(r"\w+\.\w+\(", expr_clean):
            type = "DF"
        if "lit(" in expr_clean:
            type = "DF"
        if "col(" in expr_clean:
            type = "DF"

        self.type = type

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def df_expr(self, dataframe_backend="SPARK"):
        expr = self.value
        if self.type == "SQL":
            _value = repr(self.value.replace("\n", " "))
            expr = f"F.expr({_value})"
            if dataframe_backend == "POLARS":
                expr = f"pl.Expr.laktory.sql_expr({_value})"

        return expr

    def eval(self, udfs=None, dataframe_backend=None) -> AnyDataFrameColumn:
        if dataframe_backend is None:
            dataframe_backend = settings.dataframe_backend

        # Adding udfs to global variables
        if udfs is None:
            udfs = {}
        for k, v in udfs.items():
            globals()[k] = v

        if dataframe_backend == "SPARK":
            # Imports required to evaluate expressions
            import pyspark.sql.functions as F  # noqa: F401
            import pyspark.sql.types as T  # noqa: F401
            from pyspark.sql.functions import col  # noqa: F401
            from pyspark.sql.functions import lit  # noqa: F401

        elif dataframe_backend == "POLARS":
            # Imports required to evaluate expressions
            import polars as pl  # noqa: F401
            import polars.functions as F  # noqa: F401
            from polars import col  # noqa: F401
            from polars import lit  # noqa: F401

        else:
            raise ValueError(
                f"`dataframe_backend` '{dataframe_backend}' is not supported."
            )

        _expr_str = self.df_expr(dataframe_backend=dataframe_backend)
        expr = eval(_expr_str)

        # Cleaning up global variables
        for k, v in udfs.items():
            del globals()[k]

        return expr



from pathlib import Path
from typing import Any
from typing import Literal
from typing import Union

from pydantic import AliasChoices
from pydantic import Field
from pydantic import model_validator

from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.alert import Alert
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class QueryParameterTextValue(BaseModel):
    """
    Attributes
    ----------
    value:
        Actual text value.
    """

    value: str


class QueryParameterQueryBackedValueMultiValuesOptions(BaseModel):
    """
    Attributes
    ----------
    prefix:
        Character that prefixes each selected parameter value.
    separator:
        Character that separates each selected parameter value. Defaults to a
        comma.
    suffix:
        Character that suffixes each selected parameter value.
    """

    prefix: str
    separator: str
    suffix: str


class QueryParameterQueryBackedValue(BaseModel):
    """
    Attributes
    ----------
    query_id:
        ID of the query that provides the parameter values.
    multi_values_options:
        If specified, allows multiple values to be selected for this parameter.
    values:
        List of selected query parameter values.
    """

    query_id: str = None
    multi_values_options: QueryParameterQueryBackedValueMultiValuesOptions = None
    values: list[str] = None


class QueryParameterNumericValue(BaseModel):
    """
    Attributes
    ----------
    value:
        Actual numeric value
    """

    value: float = None


class QueryParameterEnumValueMultiValuesOptions(BaseModel):
    """
    Attributes
    ----------
    prefix:
        Character that prefixes each selected parameter value.
    separator:
        Character that separates each selected parameter value. Defaults to a comma.
    suffix:
        Character that suffixes each selected parameter value.
    """

    prefix: str = None
    separator: str = None
    suffix: str = None


class QueryParameterEnumValue(BaseModel):
    """
    Attributes
    ----------
    enum_options:
        List of valid query parameter values, newline delimited.
    multi_values_options:
        If specified, allows multiple values to be selected for this parameter.
    values:
        List of selected query parameter values.
    """

    enum_options: str = None
    multi_values_options: QueryParameterEnumValueMultiValuesOptions = None
    values: list[str] = None


class QueryParameterDateValue(BaseModel):
    """
    Attributes
    ----------
    date_value:
        Manually specified date-time value
    dynamic_date_value:
        Dynamic date-time value based on current date-time
    precision:
        Date-time precision to format the value into when the query is run.
    """

    date_value: str = None
    dynamic_date_value: Literal["NOW", "YESTERDAY"] = None
    precision: Literal["DAY_PRECISION", "MINUTE_PRECISION", "SECOND_PRECISION"] = None


class QueryParameterDateRangeValueDateRangeValue(BaseModel):
    """
    Attributes
    ----------
    end:
        end of the date range
    start:
        start of the date range
    """

    end: str
    start: str


class QueryParameterDateRangeValue(BaseModel):
    """
    Attributes
    ----------
    date_range_value:
        Manually specified date-time range value
    dynamic_date_range_value:
        Dynamic date-time range value based on current date-time.
    """

    date_range_value: QueryParameterDateRangeValueDateRangeValue = None
    dynamic_date_range_value: Literal[
        "TODAY",
        "YESTERDAY",
        "THIS_WEEK",
        "THIS_MONTH",
        "THIS_YEAR",
        "LAST_WEEK",
        "LAST_MONTH",
        "LAST_YEAR",
        "LAST_HOUR",
        "LAST_8_HOURS",
        "LAST_24_HOURS",
        "LAST_7_DAYS",
        "LAST_14_DAYS",
        "LAST_30_DAYS",
        "LAST_60_DAYS",
        "LAST_90_DAYS",
        "LAST_12_MONTHS",
    ] = None
    precision: str = None
    start_day_of_week: str = None


class QueryParameter(BaseModel):
    """
    Query Parameters Specifications

    Attributes
    ----------

    name:
        Literal parameter marker that appears between double curly braces in
        the query text.
    date_range_value:
        Date-range query parameter value. Consists of following attributes
        (Can only specify one of `dynamic_date_range_value` or
        `date_range_value`)
    date_value:
        Date query parameter value. Consists of following attributes (Can only
        specify one of `dynamic_date_value` or `date_value`)
    enum_value:
        Dropdown parameter value
    numeric_value:
        Numeric parameter value
    query_backed_value:
        Query-based dropdown parameter value
    text_value:
        Text parameter value
    title:
        Text displayed in the user-facing parameter widget in the UI.
    """

    name: str
    date_range_value: QueryParameterDateRangeValue = None
    date_value: QueryParameterDateValue = None
    enum_value: QueryParameterEnumValue = None
    numeric_value: QueryParameterNumericValue = None
    query_backed_value: QueryParameterQueryBackedValue = None
    text_value: QueryParameterTextValue = None
    title: str = None


class Query(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Query

    Attributes
    ----------
    access_controls:
        Query access controls
    dirpath:
        Workspace directory inside rootpath in which the query is deployed.
        Used only if `parent_path` is not specified.
    display_name:
        Name of the query.
    query_text:
        Text of SQL query.
    warehouse_id:
        ID of a SQL warehouse which will be used to execute this query.
    apply_auto_limit:
        Whether to apply a 1000 row limit to the query result.
    catalog:
        Name of the catalog where this query will be executed.
    description:
        General description that conveys additional information about this
        query such as usage notes.
    owner_user_name:
        Query owner's username.
    parameters:
        Query parameter definition. Consists of following attributes
        (one of `*_value` is required):
    parent_path:
        The path to a workspace folder containing the query. Set to `None`
        to use user's home folder. Overwrite `rootpath` and `dirpath`. If
        changed, the query will be recreated.
    rootpath:
        Root directory to which all queries are deployed to. Can also be
        configured by settings LAKTORY_WORKSPACE_LAKTORY_ROOT environment
        variable. Default is `/.laktory/`. Used only if `parent_path` is not
        specified.
    run_as_mode:
        Sets the "Run as" role for the object.
    schema:
        Name of the schema where this query will be executed.
    tags:
        Tags that will be added to the query.

    Examples
    --------
    ```py
    from laktory import models

    query = models.resources.databricks.Query(
        display_name="google-prices",
        parent_path="/queries",
        query_text="SELECT * FROM dev.finance.slv_stock_prices",
        warehouse_id="12345",
    )
    ```
    """

    access_controls: list[AccessControl] = []
    alert: Alert = None
    apply_auto_limit: bool = None
    catalog: str = None
    description: str = None
    dirpath: str = None
    display_name: str
    name_prefix: str = None
    name_suffix: str = None
    owner_user_name: str = None
    parameters: list[QueryParameter] = None
    parent_path: Union[str, None] = None
    query_text: str
    rootpath: str = None
    run_as_mode: str = None
    schema_: str = Field(
        None, validation_alias=AliasChoices("schema", "schema_")
    )  # required not to overwrite BaseModel attribute
    tags: list[str] = None
    warehouse_id: str

    @model_validator(mode="after")
    def set_paths(self) -> Any:
        # Parent Path explicitly set
        if "parent_path" in self.model_fields_set:
            return self

        # root
        if self.rootpath is None:
            self.rootpath = settings.workspace_laktory_root

        # dir
        if self.dirpath is None:
            self.dirpath = ""
        if self.dirpath.startswith("/"):
            self.dirpath = self.dirpath[1:]

        # parent_path
        _path = Path(self.rootpath) / self.dirpath
        self.parent_path = _path.as_posix()

        return self

    @model_validator(mode="after")
    def update_name(self) -> Any:
        with self.validate_assignment_disabled():
            if self.name_prefix:
                self.display_name = self.name_prefix + self.display_name
                self.name_prefix = ""
            if self.name_suffix:
                self.display_name = self.display_name + self.name_suffix
                self.name_suffix = ""
        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.display_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        - alert
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    sql_query_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        if self.alert:
            self.alert.query_id = f"${{resources.{self.resource_name}.id}}"
            resources += [self.alert]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_renames(self) -> dict[str, str]:
        return {"schema_": "schema"}

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Query"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return [
            "access_controls",
            "alert",
            "dirpath",
            "rootpath",
            "name_prefix",
            "name_suffix",
        ]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_query"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes

    @property
    def terraform_renames(self) -> dict[str, str]:
        return self.pulumi_renames
